/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001-2002 Gran Weinholt <weinholt@linux.nu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: get.c,v 1.160 2002/07/03 07:39:33 chnix Exp $
 */
#include "giftcurs.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "transfer.h"
#include "screen.h"
#include "gift.h"
#include "get.h"
#include "poll.h"
#include "parse.h"
#include "xcommon.h"
#include "ui_transfer.h"
#include "search.h"
#include "settings.h"

#define update_ui() transfer_screen_update_downloads_if_focus()
#define DOWNLOADS (downloads.top)

/* autosearch timer interval */
#define AUTOTIMER_INTERVAL SECS(60)	/* 1 minute */
/* autosearch interval */
#define AUTOSEARCH_INTERVAL SECS(600)	/* 10 minutes */

tree downloads = TREE_INITIALIZER(transfer);

static int gift_start_download(transfer * t);
static int gift_download_addsource(transfer * t, const char *user, const char *href);
static void download_addsource(transfer * t, const char *user, const char *href);
static void new_url_handler(gift_conn * id, xml_data line, void *data);
static transfer *new_download(char *filename, unsigned long filesize, char *hash);
static int compare_source(const subhit ** p1, const subhit ** p2);
static void dl_status_handler(gift_conn * id, xml_data buf, void *data);
static void autosearch_start(void);
static void autosearch_stop(void);
static int download_search_more(transfer * d);
static void md5sum_append(char *md5file, char *hash, char *filename);

static int autotimer = 0;

int download_search(transfer * t)
{
	assert(t->hash);			/* Don't have the hash for this file */

	if (t->specific.d.search_id) {
		message(_("A source search for this file is already in progress."));
		return 0;
	}

	if (!t->id)
		transfer_reset(t);

	return download_search_more(t);
}

void download_forget(transfer * t)
{
	tree_remove_item(&downloads, t);
	transfer_destroy(t);
}

void download_delsource(transfer * t, char *href)
{
	int i;
	source *s;

	if (!(s = lookup_source(t, NULL, href))) {
		DEBUG("Tried to remove source we didn't have?");
		return;
	}

	TOUCH(t);

	for (i = 0; i < t->sourcen.num; i++) {
		if (list_index(&t->sourcen, i) == s) {
			list_remove_entry(&t->sourcen, i);
			return;
		}
	}
	abort();
}

transfer *find_download(char *hash, unsigned long size)
{
	return transfer_find(&downloads, hash, size);
}

/* Called to start a new download. */
int download_hit(hit * info)
{
	int i;
	transfer *t;
	char *md5file = get_config("set", "md5file", NULL);

	if (!info->filesize) {
		/* Downloading this file causes some really weird behaviour.
		   Btw, don't mark for translation :) */
		message("I'm sorry Dave, I'm afraid I can't do that.");
		return 0;
	}

	/* Check if we're already downloading this one. */
	if ((t = find_download(info->hash, info->filesize))) {
		if (transfer_alive(t)) {
			message(_("Already downloading '%s'."), info->filename);
			return 0;
		} else {
			download_search(t);
			message(_("Revived '%s'."), info->filename);
			return 0;
		}
	}

	/* Append the hash to a file */
	if (md5file)
		md5sum_append(md5file, info->hash, info->filename);

	t = new_download(info->filename, info->filesize, info->hash);

	if (gift_start_download(t) < 0)
		return -1;

	/* Add the sources we have so far (sorted) ... */
	qsort(info->sources.entries, info->sources.num, sizeof(subhit *), (SFunc) compare_source);
	for (i = 0; i < info->sources.num; i++) {
		subhit *sh = list_index(&info->sources, i);

		gift_download_addsource(t, sh->user, sh->href);
	}

	/* ... and search for even more :) */
	if (download_search_more(t) < 0)
		return -1;

	return 0;
}

void on_new_download(xml_data buf)
{
	/* <transfer id="7" action="download" size="183103" save="libwww-perl-5.63.tar.gz"/> */
	char *filename, *hash;
	unsigned long size;
	transfer *entry = NULL;

	if (!(filename = xml_get_param(buf, "save"))) {
		DEBUG("Download without save attribute.");
		return;
	}

	if (!(size = xml_get_param_int(buf, "size"))) {
		DEBUG("Download without size attribute.");
		return;
	}

	if (!(hash = xml_get_param(buf, "hash"))) {
		DEBUG("Download without hash attribute.");
		return;
	}

	entry = find_download(hash, size);
	if (!entry)
		entry = new_download(filename, size, hash);
	else if (entry->id)
		gift_unregister(entry->id);
	entry->id = gift_event(dl_status_handler, entry);

	update_ui();
}

void download_cleanup(void)
{
	autosearch_stop();
}


/* Callback for incoming data on id */
static void dl_status_handler(gift_conn * id, xml_data buf, void *data)
{
	transfer *t = data;
	int paused;

	assert(t->type == TRANSFER_DOWNLOAD);

	if (buf == NULL) {
		t->id = NULL;
		download_detach(t);
		message(t->transferred == t->filesize ? _("Download of %s completed.")
				: _("Download of %s remotely cancelled."), t->filename);
		if (autodestroy_all || t->autodestroy) {
			download_forget(t);
			update_ui();
		}
		return;
	}
	/* <transfer id=1 transmit=1000 total=1024/> */
	/* <transfer id="6" user="24.57.39.226" addsource="OpenFT://24.57.39.226:2768/anger.mpg"/> */
	/* <transfer id=1/> */

	if (!strcmp("chunk", xml_get_type(buf))) {
		chunk_parse(t, buf);
		return;
	}

	if (strcmp("transfer", xml_get_type(buf)))
		return;

	paused = t->paused;
	t->paused = xml_get_param_int(buf, "paused");
	if (paused != t->paused)
		transfer_touch(t);

	if (xml_get_param(buf, "addsource")) {
		char *user;

		if (!(user = xml_get_param(buf, "user"))) {
			DEBUG("Expected a user attribute.");
			return;
		}

		download_addsource(t, user, xml_get_param(buf, "addsource"));
	} else if (xml_get_param(buf, "delsource")) {
		download_delsource(t, xml_get_param(buf, "delsource"));
		update_ui();
	} else {
		t->transferred = xml_get_param_int(buf, "transmit");

		if (t->start_size == -1)
			t->start_size = t->transferred;
		t->time++;
	}
}

/* Callback for incoming data on search_id. */
static void new_url_handler(gift_conn * id, xml_data line, void *data)
{
	char *href, *user;
	transfer *t = data;

	if (!line) {
		t->specific.d.search_id = NULL;
		transfer_touch(t);
		message(_("Source search complete for '%s'."), t->filename);
		update_ui();
		return;
	}

	if (xml_get_param_int(line, "size") != t->filesize) {
		DEBUG("Dropping new source because of size mismatch.");
		return;
	}

	if (!(user = xml_get_param(line, "user"))) {
		DEBUG("Got no user tag!");
		return;
	}
	if (!(href = xml_get_param(line, "href"))) {
		DEBUG("Got no href tag!");
		return;
	}

	if (gift_download_addsource(t, user, href) >= 0)
		message(_("Got new source from %s"), user);
}

static transfer *new_download(char *filename, unsigned long filesize, char *hash)
{
	transfer *new;

	new = transfer_new(TRANSFER_DOWNLOAD, strdup(filename), filesize);
	new->id = NULL;
	new->specific.d.search_id = NULL;
	new->hash = strdup(hash);
	tree_push(&downloads, new);

	hit_download_mark(hash, filesize);

	/* haven't found a better place for this... */
	autosearch_start();
	return new;
}

static int download_search_more(transfer * t)
{
	char buf[60];

	message(_("Started a source search for '%s'."), t->hash);
	snprintf(buf, sizeof buf, "query=%s type=hash", t->hash);
	t->specific.d.search_id = gift_search_start(buf, new_url_handler, t);
	t->specific.d.latest_search = uptime();
	if (!t->specific.d.search_id)
		return -1;
	return 0;
}

static void download_addsource(transfer * t, const char *user, const char *href)
{
	source *s;

	if (lookup_source(t, NULL, href)) {
		DEBUG("Daemon tried to add a source twice.");
		return;
	}

	s = NEW(source);
	s->user = strdup(user);
	s->href = strdup(href);
	s->transmit = s->start = s->total = 0;
	list_push(&t->sourcen, s);
	TOUCH(t);
}

/*
 * This function used to just check if we already had the user, but
 * apparently you end up in the topic of #giFT for doing that.
 */
static int gift_download_addsource(transfer * t, const char *user, const char *href)
{
	source *snafu;

	if (!t->id)
		if (gift_start_download(t) < 0)
			return -1;

	if (lookup_source(t, NULL, href))
		return -2;

	if ((snafu = lookup_source(t, user, NULL)))
		if (!snafu->status || !strcmp(snafu->status, "Active") || !strcmp(snafu->status, "Queued")
			|| !strcmp(snafu->status, "Queued (Remotely)"))
			return -3;

	gift_response(t->id, "transfer user=\"%s\" addsource=\"%s\"", user, href);
	download_addsource(t, user, href);
	return 0;
}

static int gift_start_download(transfer * t)
{
	/* FIXME: should check filename for embedded " */
	if (gift_printf("<transfer action=download save=\"%s\" size=%lu hash=%s/>\n",
					t->filename, t->filesize, t->hash) < 0)
		return -1;

	t->transferred = 0;
	t->id = gift_event(dl_status_handler, t);
	return 0;
}

static int compare_source(const subhit ** p1, const subhit ** p2)
{
	const subhit *a = *p1, *b = *p2;

	if (a->availability < b->availability)
		return 1;
	if (a->availability > b->availability)
		return -1;
	return 0;
}

/* This allows people to do 'md5sum -c MD5SUMS' or something like that. */
static void md5sum_append(char *md5file, char *hash, char *filename)
{
	FILE *f;
	char *fname;

	if (md5file[0] == '/')
		/* Path is absolute */
		fname = md5file;
	else {
		/* Path is relative to completed/ */
		if (!download_dir) {
			DEBUG("Didn't have the path to the completed directory.");
			return;
		}
		mkdir(download_dir, 0755);
		fname = sprintf_alloc("%s/%s", download_dir, md5file);
	}

	f = fopen(fname, "a");

	free(fname);

	if (!f) {
		ERROR("Couldn't append md5 to '%s'", md5file);
		return;
	}

	fprintf(f, "%s  %s\n", hash, filename);

	fclose(f);
}

static int autosearch(void *data)
{
	int i, n = 0;
	tick_t now = uptime();

	assert(autotimer);
	DEBUG("Checking for downloads which need more sources...");
	for (i = 0; i < DOWNLOADS.num; i++) {
		transfer *t = list_index(&DOWNLOADS, i);

		assert(t->hash);

		if (!transfer_alive(t) || t->bandwidth)
			continue;

		if (t->specific.d.search_id)
			gift_search_stop(&t->specific.d.search_id);

		if (t->specific.d.latest_search + AUTOSEARCH_INTERVAL <= now)
			download_search(t), n++;
	}
	DEBUG("...%d downloads awakened.", n);
	return AUTOTIMER_INTERVAL;
}

static void autosearch_start(void)
{
	if (!autotimer) {
		poll_add_timer(AUTOTIMER_INTERVAL, autosearch, NULL);
		autotimer = 1;
	}
}

static void autosearch_stop(void)
{
	if (autotimer) {
		poll_del_timer(autosearch);
		autotimer = 0;
	}
}

void download_detach(transfer * t)
{
	transfer_touch(t);
	if (t->id) {
		/* FIXME: send something to giFT to cancel this transfer */
		gift_unregister(t->id);
		t->id = NULL;
	}

	list_destroy_all(&t->sourcen);

	if (t->specific.d.search_id) {
		gift_unregister(t->specific.d.search_id);
		t->specific.d.search_id = NULL;
	}
}
