/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001-2002 Gran Weinholt <weinholt@linux.nu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: gift.c,v 1.160 2002/06/30 13:14:12 chnix Exp $
 */
#include "giftcurs.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <assert.h>
#include <limits.h>

#include "screen.h"
#include "gift.h"
#include "parse.h"
#include "poll.h"
#include "xcommon.h"
#include "upload.h"
#include "ui.h"
#include "get.h"
#include "stats.h"
#include "search.h"

#include "ui_main.h"

#define RECONNECT_INTERVAL SECS(60)

struct _gift_conn {
	int id;						/* The id giFT gave us, or -1 if no yet */
	callback_func handler;		/* Called on every incoming line */
	void *data;					/* User specified data */
	list queue;					/* Lines to send when we get the id */
};

static int the_gift_fd = -1;	/* The connection we maintain with daemon */
static int pending_id = -1;		/* gift_event would like to know this */
static list id_callbacks;

static void gift_detach(void);
static int gift_handle_line(char *buf);
static int reconnect(void *data);

static gift_conn *gift_new_conn(int id)
{
	gift_conn *conn = calloc(1, sizeof *conn);

	conn->id = id;
	list_push(&id_callbacks, conn);
	return conn;
}

void gift_unregister(gift_conn * conn)
{
	int i;

	if (conn == NULL)
		return;
	i = list_find(&id_callbacks, conn);
	assert(i >= 0);

	if (conn->id < 0) {
		/* Ok, carefully now. If we haven't got the real id yet,
		 * just mark it as disabled. It will be removed when the
		 * id arrives. */
		conn->handler = NULL;
		list_free_entries(&conn->queue);
	} else {
		free(conn);
		list_remove_entry(&id_callbacks, i);
	}
}

static void gift_handle_read(int fd)
{
	static char buf[4096] = "";

	assert(fd == the_gift_fd);
	if (xgetlines(fd, buf, sizeof buf, gift_handle_line) <= 0) {
		/* shit! daemon was disconnected */
		gift_detach();
		message(_("The connection with the daemon was unexpectedly closed!"));
		ui_draw();
		/* Try to reopen with some interval */
		poll_add_timer(RECONNECT_INTERVAL, reconnect, NULL);
		return;
	}
}

int gift_attach(void)
{
	if (the_gift_fd != -1)
		return 1;
	the_gift_fd = xconnect(server_host, server_port);
	if (the_gift_fd < 0) {
		error_message(_("Can't connect to the daemon at %s:%s"), server_host, server_port);
		return -1;
	}

	poll_del_timer(reconnect);

	if (profile_name && profile_name[0])
		gift_printf("<attach profile=\"%s\" client=\"" PACKAGE
					"\" version=\"" VERSION "\"/>\n", profile_name);
	else
		gift_printf("<attach client=\"" PACKAGE "\" version=\"" VERSION "\"/>\n");
	DEBUG("Connected to %s:%s.", server_host, server_port);

	poll_add_fd(the_gift_fd, gift_handle_read);

	/* Race condition here. gift_stat_update calls gift_attach indirectly. */
	/* see note on gift_stat_update */
	/* FIXME: move to ui_main.c somehow */
	gift_stat_update(main_screen_update_stats_if_focus);

	return 0;
}

static void gift_conn_destroy(gift_conn * conn)
{
	if (conn->handler)
		conn->handler(conn, NULL, conn->data);
	free(conn);
}

void gift_detach(void)
{
	if (the_gift_fd == -1)
		return;

	list_foreach(&id_callbacks, (LFunc) gift_conn_destroy);
	list_remove_all(&id_callbacks);

	xclose(the_gift_fd);
	poll_del_fd(the_gift_fd);
	the_gift_fd = -1;
}

int gift_printf(const char *fmt, ...)
{
	va_list ap;

	if (gift_attach() < 0)
		return -1;
	va_start(ap, fmt);
	xvprintf(the_gift_fd, fmt, ap);
	va_end(ap);
	return 0;
}

int gift_response(gift_conn * conn, const char *fmt, ...)
{
	va_list ap;
	char *s;

	va_start(ap, fmt);
	s = vsprintf_alloc(fmt, ap);
	va_end(ap);

	if (conn->id < 0) {
		/* We have not got an id yet. Remember this item until then. */
		list_push(&conn->queue, s);
	} else {
		gift_printf("<%s id=%d/>\n", s, conn->id);
		free(s);
	}
	return 0;
}

gift_conn *gift_event(callback_func func, void *data)
{
	gift_conn *ret;

	ret = gift_new_conn(pending_id);
	pending_id = -1;
	ret->handler = func;
	ret->data = data;
	return ret;
}

void gift_cleanup(void)
{
	gift_detach();
	poll_del_timer(reconnect);
	gift_search_cleanup();
	download_cleanup();
}

void gift_init(void)
{
	gift_search_init();

	if (gift_attach() < 0)
		poll_add_timer(RECONNECT_INTERVAL, reconnect, NULL);
}


static int reconnect(void *data)
{
	return gift_attach() < 0 ? RECONNECT_INTERVAL : 0;
}

/* This is where we handle stuff we get from daemon. */
static int gift_handle_line(char *buf)
{
	char *head;
	int i, id;
	xml_data item;

	DEBUG("=> %s", buf);

	item = xml_parse_tag(buf);
	if (!item) {
		DEBUG("XML Parse error: %s", buf);
		return -1;
	}

	head = xml_get_type(item);
	id = xml_get_param_int(item, "id");
	if (id <= 0) {
		DEBUG("Daemon provided no id.");
		return 0;
	}

	/* find the handler for this id */
	for (i = 0; i < id_callbacks.num; i++) {
		gift_conn *cb = list_index(&id_callbacks, i);

		if (cb->id == id) {
			if (xml_param_count(item) == 1) {
				gift_conn_destroy(cb);
				list_remove_entry(&id_callbacks, i);
			} else
				cb->handler(cb, item, cb->data);
			return 0;
		}
		if (cb->id < 0) {
			/* This callback is in turn to receive its id.
			 * Check if we should catch this or let it go */
			if (strcmp(head, "event"))
				continue;
			/* Flush the send queue */
			for (i = 0; i < cb->queue.num; i++)
				gift_printf("<%s id=%d/>\n", (char *) cb->queue.entries[i], id);
			list_free_entries(&cb->queue);
			/* see comment on gift_unregister_id above */
			if (!cb->handler) {
				gift_conn_destroy(cb);
				list_remove_entry(&id_callbacks, i);
			} else {
				cb->id = id;
			}
			return 0;
		}
	}

	if (!strcmp(head, "transfer")) {
		/* upload has format: <transfer id="2" action="upload" size=... */
		/* download should have the same */
		char *action = xml_get_param(item, "action");

		/* Set this so that these callbacks can call gift_event if they
		 * want to establish a connection */
		pending_id = id;
		if (action) {
			if (!strcasecmp(action, "upload"))
				on_new_upload(item);
			else if (!strcasecmp(action, "download"))
				on_new_download(item);
		}
		pending_id = -1;
		return 0;
	}

	/* Note: Right now it's normal to get here, because the daemon
	 * happily spits out new results for cancelled searches. There is
	 * yet no way to tell it that we don't want any more.
	 */
	DEBUG("No handler for id %d.", id);
	return 0;
}
