/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001-2002 Gran Weinholt <weinholt@linux.nu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: list.c,v 1.53 2002/06/17 17:28:25 weinholt Exp $
 */
#include "giftcurs.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "list.h"
#include "screen.h"
#include "parse.h"

int SCROLLOFF = 0;

static void list_resize(list * snafu, int nmemb);

void list_initialize(list * snafu)
{
	memset(snafu, 0, sizeof(list));
}

list *list_new(void)
{
	list *foo = malloc(sizeof(list));

	if (!foo)
		FATAL("malloc");
	list_initialize(foo);
	return foo;
}

void list_remove_all(list * snafu)
{
	snafu->num = 0;
	list_resize(snafu, 0);
	snafu->start = 0;
	snafu->sel = -1;
}

void list_free_entries(list * snafu)
{
	int i;

	for (i = 0; i < snafu->num; i++) {
		free(snafu->entries[i]);
		snafu->entries[i] = NULL;
	}
	list_remove_all(snafu);
}

void list_push(list * snafu, void *entry)
{
	if (snafu->allocated <= snafu->num)
		list_resize(snafu, 1 + snafu->num * 2);

	snafu->entries[snafu->num++] = entry;
}

void list_insert(list * snafu, void *entry, int pos)
{
	if (snafu->allocated <= snafu->num)
		list_resize(snafu, 1 + snafu->num * 2);

	memmove(snafu->entries + pos + 1, snafu->entries + pos, (snafu->num - pos) * sizeof(void *));
	snafu->num++;
	snafu->entries[pos] = entry;
}

void list_push_first(list * snafu, void *entry)
{
	list_insert(snafu, entry, 0);
}

void list_remove_entry(list * snafu, int index)
{
	assert(0 <= index && index < snafu->num);
	if (snafu->num == 1) {
		list_remove_all(snafu);
	} else {
		void **base = snafu->entries + index;

		memmove(base, base + 1, (--snafu->num - index) * sizeof base[0]);
		if (snafu->allocated > 4 * snafu->num)
			/* a quarter is used, shrink the size to the half */
			list_resize(snafu, 2 * snafu->num);
	}
}

void list_resize(list * snafu, int nmemb)
{
	if (!nmemb) {
		free(snafu->entries);
		snafu->entries = NULL;
	} else if (!snafu->allocated)
		snafu->entries = malloc(sizeof(void *) * nmemb);
	else
		snafu->entries = realloc(snafu->entries, sizeof(void *) * nmemb);

	snafu->allocated = nmemb;
}

void list_check_values(list * snafu, int h)
{
	int scrolloff = SCROLLOFF;

	/* Do sanity check on values */
	if (snafu->start < 0)
		snafu->start = 0;
	if (snafu->start > snafu->num - 1)
		snafu->start = snafu->num - 1;
	if (snafu->sel < 0)
		snafu->sel = 0;
	if (snafu->sel > snafu->num - 1)
		snafu->sel = snafu->num - 1;

	/* Check against list frame */
	if (h <= SCROLLOFF * 2)
		scrolloff = (h - 1) / 2;
	/* check against bottom */
	if (snafu->start < snafu->sel - h + 1 + scrolloff)
		snafu->start = snafu->sel - h + 1 + scrolloff;
	if (snafu->start > snafu->num - h)
		snafu->start = snafu->num - h;
	/* check against top */
	if (snafu->start > snafu->sel - scrolloff)
		snafu->start = snafu->sel - scrolloff;
	if (snafu->start < 0)
		snafu->start = 0;
}

void *list_selected(list * snafu)
{
	return snafu->num ? snafu->entries[snafu->sel] : NULL;
}

/* Call this if an entry has changed and should be sorted to a lower index.
 * Keeps track of lower and upper bound, always halving the possible range.
 * Therefore it's O(log n) */
void list_resort(list * snafu, int pos, SFunc cmp)
{
	int lower = 0, upper = pos;
	void *item = snafu->entries[pos];

	/* this function only sorts in one direction */
	assert(pos >= snafu->num - 1 || cmp(item, snafu->entries[pos + 1]) <= 0);

	while (upper > lower) {
		int mid = (upper + lower) / 2;
		int res = cmp(item, snafu->entries[mid]);

		if (res <= 0)
			upper = mid;
		if (res >= 0)
			lower = mid + 1;
	}

	memmove(snafu->entries + lower + 1, snafu->entries + lower, (pos - lower) * sizeof(void *));
	snafu->entries[lower] = item;

	if (snafu->sel == pos)
		snafu->sel = lower;
	else if (lower <= snafu->sel && snafu->sel < pos)
		snafu->sel++;
}

/* inserts an item into a sorted list at the right place. */

void list_insort(list * snafu, void *item, SFunc cmp)
{
	list_push(snafu, item);
	list_resort(snafu, snafu->num - 1, cmp);
}

int list_lookup(const char *needle, list * snafu)
{
	return lookup(needle, (const char **) snafu->entries, snafu->num);
}

void list_foreach(list * snafu, LFunc func)
{
	int i;

	for (i = 0; i < snafu->num; i++)
		func(snafu->entries[i]);
}

int list_find(list * snafu, void *elem)
{
	int i;

	for (i = 0; i < snafu->num; i++)
		if (snafu->entries[i] == elem)
			return i;
	return -1;
}
