/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001-2002 Gran Weinholt <weinholt@linux.nu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: parse.c,v 1.78 2002/06/02 17:10:42 weinholt Exp $
 */
#include "giftcurs.h"

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>

#include "parse.h"
#include "screen.h"

#if UNENTITY
static char *unentity_one(char *dst, char *src)
{
	/* &#466; &gt; &#x7F; */
	static const char *entities[] = { "amp", "lt", "gt", "nbsp", "quot", "apos" };
	static const char result[] = /*"&<> \"'" */ "&<> ''";
	char *end;

	src++;

	if ((end = strchr(src, ';'))) {
		if (*src == '#') {
			/* numerical */
			char *endp;
			int val;

			if (tolower(src[1]) == 'x')
				val = strtoul(src + 2, &endp, 16);
			else
				val = strtoul(src + 1, &endp, 10);
			if (endp == end) {
				*dst = (char) val;
				return end + 1;
			}
		} else {
			/* literal */
			int i;

			for (i = 0; i < buflen(entities); i++) {
				int len = strlen(entities[i]);

				if (!strncmp(src, entities[i], len)) {
					*dst = result[i];
					return end + 1;
				}
			}
		}
	}
	*dst = '&';
	return src;
}
#endif

xml_data xml_parse_tag(char *str)
{
	/* insert null charachters at appropriate places */
	/* <giFT VERSION=1 NASSE="mkt litet djur"/> becomes */
	/* giFT\0VERSION=1\0NASSE=mkt litet djur\0\0 */
	char *dst, *src;

	if (*str != '<')
		return NULL;
	src = ++str;
	do {
		if (*src == '\0' || *src == '>')
			return NULL;
		src++;
	} while (*src != ' ' && *src != '/');
	/* got past the header */
	dst = src;
	for (;;) {
		char c;

		/* check for end of tag */
		*dst++ = '\0';
		if (*src++ == '/') {
			if (*src != '>')
				return NULL;
			*dst = '\0';
			return str;
		}
		/* find equal sign */
		do {
			if (*src == '\0')
				return NULL;
			*dst++ = *src;
		} while (*src++ != '=');

		/* isolate the value. With or without quotes */
		c = *src;
		if (c == '"' || c == '\'') {
			src++;
			while (*src != c) {
				if (*src == '\0')
					return NULL;
#if UNENTITY
				if (*src == '&')
					src = unentity_one(dst, src);
				else
#endif
					*dst = *src++;
				dst++;
			}
			src++;
		} else {
			do {
				if (*src == '\0')
					return NULL;
#if UNENTITY
				if (*src == '&')
					src = unentity_one(dst, src);
				else
#endif
					*dst = *src++;
				dst++;
			} while (*src != ' ' && *src != '/');
		}
	}
}

int xml_param_count(xml_data str)
{
	int i = 0;

	str = strchr(str, '\0') + 1;
	while (*str) {
		str = strchr(str, '\0') + 1;
		i++;
	}
	return i;
}

char *xml_get_param(xml_data str, const char *prop)
{
	int len = strlen(prop);

	str = strchr(str, '\0') + 1;
	while (*str) {
		char *end = strchr(str, '=');

		if (end - str == len && memcmp(str, prop, len) == 0)
			return end + 1;
		str = strchr(end, '\0') + 1;
	}
	return NULL;
}

unsigned long xml_get_param_int(xml_data str, const char *prop)
{
	char *foo;
	char *endp;

	foo = xml_get_param(str, prop);
	if (foo) {
		unsigned long val = strtoul(foo, &endp, 10);

		if (!endp[0])
			return val;
	}
	return 0;
}

char *http_decode(char *src)
{
	char *dst;
	char *fnord = src;

	for (dst = src; *src; src++, dst++) {
		if (*src == '+')
			*dst = ' ';
		else if (*src == '%') {
			char evil;

			src += 3;
			evil = *src;
			*src = '\0';
			*dst = strtoul(src - 2, NULL, 16);
			*src = evil;
			src--;
			continue;
		}
		*dst = *src;
	}
	*dst = '\0';
	return fnord;
}

/* heavily optimized, and thus less readable */
char *trim(char *src)
{
	char *dst = src, *last = src + 666666;
	char *fnord = src;

	while (*src) {
		if (!isspace(*src)) {
			if (src > last)
				*dst++ = ' ';
			*dst++ = *src++;
			last = src;
		} else
			src++;
	}
	*dst = '\0';

	return fnord;
}

void strpad(char *s, int length)
{
	char *p = strchr(s, '\0');
	int i = length - (p - s);

	if (i > 0)
		memset(p, ' ', i);
	p[i] = '\0';
}

char *humanify(unsigned long quantity)
{
	return humanify_scale(quantity, 0);
}

char *humanify_scale(unsigned long quantity, int steps)
{
	const char *suffix = "bkMGTPEZY";
	static char ringbuf[5][5];
	static int index;
	static char point = 0;
	char *buf = ringbuf[index++];

	/* Get the locale decimal point. */
	if (point == 0) {
		char tmp[5];

		snprintf(tmp, 4, "%.1f", 0.0);
		point = tmp[1];
	}

	if (index == buflen(ringbuf))
		index = 0;

	/*
	   if (quantity == 0) {
	   strcpy(buf, "   0");
	   return buf;
	   }
	 */
	if (steps == 1 && quantity < 10) {
		/* convert 3k => 3072 */
		steps--;
		quantity *= 1024;
	}
	if (steps == 0 && quantity < 10000) {
		sprintf(buf, "%4d", (int) quantity);
		return buf;
	}
	while (quantity >= 10240) {
		steps++;
		quantity /= 1024;
	}
	if (quantity < 1000) {
		sprintf(buf, "%3d%c", (int) quantity, suffix[steps]);
	} else {
		quantity = quantity * 10 / 1024;
		/* quantity now in range 09-99, make a decimal point */
		sprintf(buf, " %02d%c", (int) quantity, suffix[steps + 1]);
		buf[0] = buf[1];
		buf[1] = point;
	}
	return buf;
}

char *humanify_time(unsigned long secs)
{
	/* 34d12h, 59m59s ... */
	static char ringbuf[2][6];
	static int index;
	char *buf = ringbuf[index++];
	int minutes, hours, days;

	if (index < buflen(ringbuf))
		index = 0;

	if (secs == 0) {
		strcpy(buf, "0s");
		return buf;
	}
	if (secs < 60) {
		sprintf(buf, "%ds", (int) secs);
		return buf;
	}
	minutes = secs / 60;
	if (minutes < 60) {
		sprintf(buf, "%dm%ds", minutes, (int) secs % 60);
		return buf;
	}
	hours = minutes / 60;
	if (hours < 24) {
		sprintf(buf, "%dh%dm", hours, minutes % 60);
		return buf;
	}
	days = hours / 24;
	if (days < 100)
		sprintf(buf, "%dd%dh", days, hours % 24);
	else
		snprintf(buf, sizeof ringbuf[0], "%dd", days);
	return buf;
}

#if !HAVE_STRNDUP
char *strndup(const char *s, size_t size)
{
	int n = strnlen(s, size);
	char *tmp = malloc(n + 1);

	if (tmp) {
		memcpy(tmp, s, n);
		tmp[n] = '\0';
	}
	return tmp;
}
#endif

#if !HAVE_STRNLEN
size_t strnlen(const char *s, size_t maxlen)
{
	const char *end = memchr(s, '\0', maxlen);

	return end ? end - s : maxlen;
}
#endif

int lookup(const char *needle, const char *haystack[], int items)
{
	int i;

	for (i = 0; i < items; i++)
		if (haystack[i] && !strcmp(needle, haystack[i]))
			return i;
	return -1;
}

int strncatf(char *buf, int max, const char *fmt, ...)
{
	va_list ap;
	int len = strlen(buf);

	va_start(ap, fmt);
	len = vsnprintf(buf + len, max - len, fmt, ap);
	va_end(ap);
	return len;
}

#if (VA_COPY == use_assign_it_works)
# undef VA_COPY
# define VA_COPY(dst, src) dst = src
#endif

char *vsprintf_alloc(const char *fmt, va_list ap)
{
	/* Partly stolen from the Linux manpage for vsprintf */
	int size = 1024;
	char *p;

	for (;;) {
		int n;
		va_list aq;

		p = malloc(size);
		if (!p)
			return NULL;
		/* Try to print in the allocated space. */
		VA_COPY(aq, ap);
		n = vsnprintf(p, size, fmt, aq);
		va_end(aq);
		/* If that worked, break the loop. */
		if (n > -1 && n < size)
			return realloc(p, n + 1);
		/* Else try again with more space. */
		if (n > -1)				/* glibc 2.1 */
			size = n + 1;		/* precisely what is needed */
		else					/* glibc 2.0 */
			size *= 2;			/* twice the old size */
		free(p);
	}
}

char *sprintf_alloc(const char *fmt, ...)
{
	va_list ap;
	char *ret;

	va_start(ap, fmt);
	ret = vsprintf_alloc(fmt, ap);
	va_end(ap);
	return ret;
}

void insert_at_ratio(char *buf, char *ins, float ratio, int max)
{
	int pos;
	int len = strlen(ins);

	ratio = ratio < 0 ? 0 : ratio > 1 ? 1 : ratio;
	pos = ratio * max;
	memmove(buf + pos + len, buf + pos, strlen(buf) - pos);
	memcpy(buf + pos, ins, len);
}

char *stristr(const char *haystack, const char *needle)
{
	char *lower, *ret;
	int len;

	/* If all-lowercase, ignore case. Otherwise match case. */
	ret = strstr(haystack, needle);
	if (ret)
		return ret;

	len = strlen(haystack) + 1;
	lower = malloc(len);
	while (len--)
		lower[len] = tolower(haystack[len]);
	ret = strstr(lower, needle);
	free(lower);
	if (ret)
		return haystack - lower + ret;
	return NULL;
}

char *remove_word(char *foo, int *posp)
{
	int i = *posp;

	while (foo[i - 1] == ' ' && i > 0)
		i--;
	while (foo[i - 1] != ' ' && i > 0)
		i--;
	strcpy(foo + i, foo + *posp);
	*posp = i;

	return foo;
}
