/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001-2002 Gran Weinholt <weinholt@linux.nu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: poll.c,v 1.46 2002/06/06 16:23:08 chnix Exp $
 */
#include "giftcurs.h"

#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/poll.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>
#include <assert.h>

#include "screen.h"
#include "poll.h"
#include "list.h"

#define MAX_FDS 4

struct timer {
	tick_t at;
	TFunc fn;
	void *data;
};

static int compare_timers(struct timer *a, struct timer *b);
static void poll_once(void);

/* Filedescriptors to be monitored are stored in two parallel arrays */
static struct pollfd pfds[MAX_FDS];
static FFunc fd_handlers[MAX_FDS];
static int nfds = 0;

/* Timers are stored in a sorted list */
static list timers;

/* Timekeeping stuff */
static tick_t uptime_cached = 0;
static time_t basetime = 0;

int poll_add_timer(unsigned long delay, TFunc fn, void *data)
{
	struct timer *new = malloc(sizeof *new);

	new->at = uptime() + delay;
	new->fn = fn;
	new->data = data;
	list_insort(&timers, new, (SFunc) compare_timers);
	return 0;
}

int poll_del_timer(TFunc fn)
{
	int i;

	for (i = 0; i < timers.num; i++) {
		struct timer *t = list_index(&timers, i);

		if (t->fn == fn) {
			free(t);
			list_remove_entry(&timers, i);
			return 0;
		}
	}
	return -1;
}

int poll_add_fd(int fd, FFunc fn)
{
	int i;

	assert(fn != NULL);

	for (i = 0; i < nfds; i++)
		if (pfds[i].fd == fd) {
			DEBUG("Reusing fd %d.", fd);
			if (fd_handlers[i] != fn) {
				DEBUG("Changing handler function for fd %d.", fd);
				fd_handlers[i] = fn;
			}
			return 0;
		}
	assert(nfds < MAX_FDS);
	pfds[nfds].fd = fd;
	pfds[nfds].events = POLLIN;
	fd_handlers[nfds++] = fn;
	return 0;
}

int poll_del_fd(int fd)
{
	int i;

	for (i = 0; pfds[i].fd != fd; i++)
		if (i >= nfds)
			return -1;
	pfds[i] = pfds[--nfds];
	fd_handlers[i] = fd_handlers[nfds];
	return 0;
}

void timer_destroy(void)
{
	assert(nfds == 0);
	assert(timers.num == 0);
}

void poll_forever(void)
{
	while (nfds || timers.num)
		poll_once();
}

/* Wait for an event to occur and handle it */
static void poll_once(void)
{
	int ev, fd, i, timeout = -1;
	struct timer *t = NULL;

	if (timers.num) {
		t = list_index(&timers, 0);

		timeout = t->at - uptime();
		if (timeout <= 0)
			goto dispatch_timer;
		timeout *= 10;			/* convert from centiseconds to milliseconds */
	}

	/* This is the only place where we sleep - invalidate uptime cache */
	uptime_cached = 0;

	switch (poll(pfds, nfds, timeout)) {
	case -1:
		if (errno != EINTR)
			ERROR("poll");
		return;
	case 0:
	  dispatch_timer:
		list_remove_entry(&timers, 0);

		timeout = t->fn(t->data);

		if (timeout) {
			t->at = uptime() + timeout;
			list_insort(&timers, t, (SFunc) compare_timers);
		} else
			free(t);
		return;
	default:
		assert(nfds);
		/* Find the first guilty */
		for (i = 0; (ev = pfds[i].revents) == 0; i++)
			assert(i < nfds);

		/* chech for errors we did */
		if (ev & POLLNVAL || ev & POLLOUT)
			abort();
		fd = pfds[i].fd;

		/* chech for errors they did */
		if (ev & POLLERR || ev & POLLHUP) {
			DEBUG("POLL(ERR|HUP) on fd %d.", fd);
			close(fd);
		} else
			assert(ev & POLLIN);

		fd_handlers[i] (fd);
		return;
	}
}

/* calculates number of ticks since program start */
tick_t uptime(void)
{
	struct timeval tv;

	/* caching of uptime value cuts number of gettimeofday calls by half */
	if (uptime_cached)
		return uptime_cached;

	assert(basetime);
	gettimeofday(&tv, NULL);
	/* read the time in centiseconds */
	uptime_cached = (tv.tv_sec - basetime) * 100 + tv.tv_usec / 10000;
	return uptime_cached;
}

static int compare_timers(struct timer *a, struct timer *b)
{
	if (a->at < b->at)
		return -1;
	if (a->at > b->at)
		return +1;
	return 0;
}

void timer_init(void)
{
	time(&basetime);
}
