/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001-2002 Gran Weinholt <weinholt@linux.nu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: search.c,v 1.57 2002/07/07 17:38:07 weinholt Exp $
 */
#include "giftcurs.h"

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#define __USE_GNU
#include <string.h>
#include <ctype.h>

#include "screen.h"
#include "gift.h"
#include "parse.h"
#include "poll.h"
#include "xcommon.h"
#include "ui.h"
#include "ui_main.h"			/* ugly */
#include "get.h"
#include "search.h"
#include "ban.h"
#include "settings.h"

#define HITS_INTERVAL SECS(0.5)

#define update_ui() incoming_search_item()
#define HITS(q) ((q)->hits.top)

list queries = LIST_INITIALIZER;

/*
 * There a two types of searches, namely searches for text (`query') and
 * searches for hash (in get.c)
 */
static void gift_query_parse(gift_conn *, xml_data, void *data);
static int qsort_comparer(hit ** p1, hit ** p2);
static int compare_sources_filesize(const hit * h1, const hit * h2);
static int compare_availability_filesize(const hit * h1, const hit * h2);
static int compare_filesize(const hit * h1, const hit * h2);
static int compare_alfabet(const hit * h1, const hit * h2);
static int compare_alfabetpath(const hit * h1, const hit * h2);
static void add_subhit(hit * bob, char *user, char *href, int av);
static int gift_query_timer(void *data);

static SFunc hit_compare = (SFunc) compare_availability_filesize;
static int max_hits = 1000;
static int filter_hits = 0;

int gift_query_start(char *searchstr)
{
	query *q = list_selected(&queries);

	q->id = gift_search_start(searchstr, gift_query_parse, q);
	if (!q->id)
		return -1;
	q->date = uptime();
	return 0;
}

gift_conn *gift_search_start(char *searchstr, callback_func fn, void *data)
{
	if (gift_printf("<search %s/>\n", searchstr) < 0)
		return NULL;
	return gift_event(fn, data);	/* fn gets called on next <event id=%d> */
}

void gift_search_stop(gift_conn ** id)
{
	/* FIXME: tell giFT to stop the search */
	gift_unregister(*id);
	*id = NULL;
}


void query_destroy(query * q)
{
	free(q->search_term);
	gift_query_clear(q);
	free(q);
}

void gift_search_cleanup(void)
{
	int i;

	poll_del_timer(gift_query_timer);
	for (i = 0; i < queries.num; i++)
		query_destroy(list_index(&queries, i));
	list_remove_all(&queries);
	banwords_cleanup();
}

void gift_search_init(void)
{
	query *q;

	max_hits = atoi(get_config("set", "max-hits", "1000"));
	filter_hits = atoi(get_config("set", "filter-hits", "0"));

	list_initialize(&queries);
	/* Add a dummy search, which represents empty search fields. */
	q = calloc(1, sizeof(query));
	q->search_term = strdup("");
	q->realm = 0;
	q->id = NULL;
	tree_initialize(&q->hits, &hit_methods);
	list_push(&queries, q);

	poll_add_timer(HITS_INTERVAL, gift_query_timer, NULL);
	banwords_init();
}

char *extract_directory_alloc(const char *href)
{
	/* This is just a hack to get at least _something_ in the infobox
	 * Some examples:
	 * OpenFT://123.34.5.99:1234/ _directory_ /file.gz
	 * OpenFT://a.b.c.d:e@x.y.z.p:q/ _direcory_ /foo.tar
	 * Note: We should not depend on the protocol-specific part
	 */
	char *end, *s = strchr(href, '.');

	if (!s)
		return NULL;
	s = strchr(s, '/');
	if (!s)
		return NULL;
	s++;
	end = strrchr(s, '/');
	if (!end)
		return NULL;
	s = strndup(s, end - s);
	http_decode(s);
	trim(s);
	/* is it worth to shrink the allocated memory? */
	return realloc(s, strlen(s) + 1);
}

char *extract_filename_alloc(const char *href)
{
	char *buf = strrchr(href, '/');
	char *buf2 = strrchr(href, '\\');

	if (buf2 > buf)
		buf = buf2;
	/* show something even if there are no file (e.g. OpenFT nodepage) */
	buf = strdup(buf[1] ? buf + 1 : "[index]");
	http_decode(buf);
	trim(buf);
	/* is it worth to shrink the allocated memory? */
	buf = realloc(buf, strlen(buf) + 1);
	/* and replace " with ' because we cannot escape it yet */
	for (buf2 = buf; *buf2; buf2++)
		if (*buf2 == '"')
			*buf2 = '\'';
	return buf;
}

void gift_query_stop(query * q)
{
	gift_search_stop(&q->id);
/*	if (q->id) {
		gift_unregister(q->id);
		q->id = NULL;
	}*/
}

void gift_query_clear(query * q)
{
	tree_destroy_all(&q->hits);
	HITS(q).sel = -1;
	HITS(q).start = 0;
}

void subhit_destroy(subhit * sh)
{
	free(sh->user);
	free(sh->href);
	free(sh->pretty);
	free(sh);
}

void hit_destroy(hit * bob)
{
	free(bob->pretty);
	free(bob->filename);
	free(bob->directory);
	free(bob->hash);
	list_destroy_all(&bob->sources);
	free(bob);
}

char *change_sort_method(int direction)
{
	/* direction == 0 just sorts with current method. */
	/* direction == -1 or +1 changes method and sorts */
	static const void *methods[] = { compare_sources_filesize, compare_availability_filesize,
		compare_filesize, compare_alfabet, compare_alfabetpath, NULL
	};
	int i;
	query *q = list_selected(&queries);

	/* Change sorting method */
	for (i = 0; methods[i]; i++)
		if (methods[i] == hit_compare)
			break;
	if ((i += direction) < 0)
		i = buflen(methods) - 2;
	if (!(hit_compare = methods[i]))
		hit_compare = methods[0];

	tree_sort(&q->hits, (SFunc) qsort_comparer);

	update_ui();

	if (hit_compare == (SFunc) compare_alfabet)
		return _("alphabetical");
	if (hit_compare == (SFunc) compare_alfabetpath)
		return _("alphabetical by path");
	if (hit_compare == (SFunc) compare_filesize)
		return _("size-based");
	if (hit_compare == (SFunc) compare_availability_filesize)
		return _("availability- and size-based");

	return _("number of sources- and size-based");
}

void hit_download_mark(const char *hash, unsigned long size)
{
	int i, j;

	for (j = 0; j < queries.num; j++) {
		query *q = list_index(&queries, j);

		for (i = 0; i < HITS(q).num; i++) {
			hit *h = list_index(&HITS(q), i);

			if (h->filesize == size && !strcmp(hash, h->hash)) {
				h->downloading = 1;
				/* invalidate previous pretty-string */
				TOUCH(h);
				return;
			}
		}
	}
}

/* Update the UI with new hits periodically. */
static int gift_query_timer(void *data)
{
	if (queries.num) {
		query *q = list_selected(&queries);

		if (q->dirty)
			update_ui();
	}
	return HITS_INTERVAL;
}

static int qsort_comparer(hit ** p1, hit ** p2)
{
	return hit_compare(*p1, *p2);
}

/* sort against number of sources and filesize */
static int compare_sources_filesize(const hit * h1, const hit * h2)
{
	if (h1->sources.num < h2->sources.num)
		return 1;
	if (h1->sources.num > h2->sources.num)
		return -1;
	if (h1->filesize < h2->filesize)
		return 1;
	if (h1->filesize > h2->filesize)
		return -1;
	return 0;
}

/* sort against availability and filesize */
static int compare_availability_filesize(const hit * h1, const hit * h2)
{
	int i, a1 = 0, a2 = 0;

	for (i = 0; i < h1->sources.num; i++)
		a1 += !!((subhit *) list_index(&h1->sources, i))->availability;
	for (i = 0; i < h2->sources.num; i++)
		a2 += !!((subhit *) list_index(&h2->sources, i))->availability;

	if (a1 < a2)
		return 1;
	if (a1 > a2)
		return -1;

	if (h1->sources.num < h2->sources.num)
		return 1;
	if (h1->sources.num > h2->sources.num)
		return -1;

	if (h1->filesize < h2->filesize)
		return 1;
	if (h1->filesize > h2->filesize)
		return -1;
	return 0;
}

/* sort against filesize */
static int compare_filesize(const hit * h1, const hit * h2)
{
	if (h1->filesize < h2->filesize)
		return 1;
	if (h1->filesize > h2->filesize)
		return -1;
	return 0;
}

/* sort alphabetically */
static int compare_alfabet(const hit * h1, const hit * h2)
{
	return strcoll(h1->filename, h2->filename);
}

/* sort alphabetically by path */
static int compare_alfabetpath(const hit * h1, const hit * h2)
{
	subhit *s1, *s2;
	char *p1, *p2;

	s1 = (subhit *) list_index(&h1->sources, 0);
	s2 = (subhit *) list_index(&h2->sources, 0);

	p1 = strchr(strchr(s1->href, ':') + 3, '/');
	p2 = strchr(strchr(s2->href, ':') + 3, '/');

	if (!p1 || !p2) {
		/* XXX: One of these urls is kinda b0rken... */
		return strcoll(s1->href, s2->href);
	}

	return strcoll(p1 + 1, p2 + 1);
}

static void add_subhit(hit * bob, char *user, char *href, int av)
{
	subhit *sh = NEW(subhit);

	sh->user = strdup(user);
	sh->href = strdup(href);
	sh->availability = av;
	list_push(&bob->sources, sh);
}

/* This is the big ugly function that parses search hits */
static void gift_query_parse(gift_conn * id, xml_data buf, void *data)
{
	hit *bob;
	int i;
	char *href, *hash, *user;
	unsigned long filesize;
	query *q = data;

	if (!buf) {
		/* end of search */
		q->id = NULL;
		if ( /*q->dirty && */ q == list_index(&queries, queries.sel))
			update_ui();
		message(ngettext("Search complete. We got %i unique hit for %s.",
						 "Search complete. We got %i unique hits for %s.",
						 HITS(q).num), HITS(q).num, q->search_term);
		return;
	}

	href = xml_get_param(buf, "href");
	if (!href || !*href || !strchr(href, ':')) {
		DEBUG("Malformed URL.");
		return;
	}

	if (!banwords_check(href))
		return;

	hash = xml_get_param(buf, "hash");
	filesize = xml_get_param_int(buf, "size");
	user = xml_get_param(buf, "user");

	if (!user || !hash)
		return;

	if (filter_hits)
		if (!(xml_get_param_int(buf, "availability")))
			return;

	/* Check if we already had this one. */
	for (i = 0; i < HITS(q).num; i++) {
		int j;

		bob = list_index(&HITS(q), i);

		if (strcmp(bob->hash, hash) || bob->filesize != filesize)
			continue;

		/* Check if we already have this user */
		for (j = 0; j < bob->sources.num; j++) {
			subhit *sh = list_index(&bob->sources, j);

			if (!strcmp(user, sh->user))
				return;
		}

		/* Never seen this href before */
		add_subhit(bob, user, href, xml_get_param_int(buf, "availability"));

		/* Invalidate previous pretty-string */
		TOUCH(bob);
		if (q == list_selected(&queries))
			list_resort(&HITS(q), i, (SFunc) hit_compare);
		q->dirty = 1;
		return;
	}

	if (max_hits) {
		if (HITS(q).num == max_hits - 1)
			message(_("The limit of %d unique hits was reached."), max_hits);
		else if (HITS(q).num >= max_hits)
			return;
	}

	/* It's a brand new hit */
	bob = NEW(hit);
	bob->filesize = filesize;
	bob->hash = strdup(hash);
	bob->filename = extract_filename_alloc(href);;
	bob->directory = extract_directory_alloc(href);
	list_initialize(&bob->sources);
	add_subhit(bob, user, href, xml_get_param_int(buf, "availability"));
	bob->downloading = !!find_download(hash, filesize);

	if (q == list_selected(&queries))
		list_insort(&HITS(q), bob, (SFunc) hit_compare);
	q->dirty = 1;
}
