/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001-2002 Gran Weinholt <weinholt@linux.nu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: transfer.c,v 1.45 2002/06/19 23:14:43 weinholt Exp $
 */
#include "giftcurs.h"

#include <stdlib.h>
#include <string.h>

#include "get.h"
#include "upload.h"
#include "transfer.h"
#include "gift.h"
#include "screen.h"
#include "parse.h"

/* This should never be used. It's only here so it can be translated. */
static __attribute__ ((unused))
const char *_statuses[] = {
	N_("Waiting"),
	N_("Paused"),
	N_("Queued (Remotely)"),
	N_("Queued"),
	N_("Active"),
	N_("Complete"),
	N_("Cancelled (Remotely)"),
	N_("Timed out")
};

int autodestroy_all = 0;

static int transfer_calculate(transfer * t);
static unsigned long bandwidth_shown(transfer * t);

void source_destroy(source * s)
{
	free(s->user);
	free(s->href);
	free(s->pretty);
	free(s->status);
	free(s);
}

void source_touch(source * s)
{
	TOUCH(s);
}

void transfer_destroy(transfer * t)
{
	transfer_detach(t);
	free(t->filename);
	free(t->pretty);
	free(t->hash);
	free(t);
}

void transfer_touch(transfer * t)
{
	TOUCH(t);
	if (t->type == TRANSFER_DOWNLOAD)
		list_foreach(&t->sourcen, (LFunc) source_touch);
}

void transfer_detach(transfer * t)
{
	transfer_touch(t);

	t->paused = 0;

	list_destroy_all(&t->sourcen);

	if (t->type == TRANSFER_DOWNLOAD) {
		download_detach(t);
	} else {
		upload_detach(t);
	}
}

transfer *transfer_new(int type, char *name, unsigned long size)
{
	transfer *t = NEW(transfer);

	t->id = NULL;
	t->filename = name;
	t->filesize = size;
	t->start_size = -1;
	t->bandwidth = -1;
	t->type = type;
	list_initialize(&t->sourcen);
	transfer_reset(t);
	return t;
}

void transfer_cancel(transfer * t)
{
	if (t->id)
		gift_response(t->id, "transfer action=cancel");

	transfer_detach(t);
}

void transfer_reset(transfer * t)
{
	/* reset progress statistics */
	t->transferred = t->bandwidth = -1;
}

void transfer_suspend(transfer * t)
{
	if (t->paused)
		gift_response(t->id, "transfer action=unpause");
	else
		gift_response(t->id, "transfer action=pause");
}

transfer *transfer_find(tree * tr, char *hash, unsigned long size)
{
	int i;

	for (i = 0; i < tr->top.num; i++) {
		transfer *t = list_index(&tr->top, i);

		if (t->filesize == size && !strcmp(t->hash, hash))
			return t;
	}
	return NULL;
}

transfer *transfer_find_dead(tree * tr, char *hash, unsigned long size)
{
	int i;

	for (i = 0; i < tr->top.num; i++) {
		transfer *t = list_index(&tr->top, i);

		if (t->filesize == size && !strcmp(t->hash, hash) && !transfer_alive(t))
			return t;
	}
	return NULL;
}

/* Returns 1 if at least one transfer has changed */
int transfers_calculate(void)
{
	int i, fnord = 0;

	for (i = 0; i < downloads.top.num; i++)
		if (transfer_calculate(list_index(&downloads.top, i))) {
			transfer_touch(list_index(&downloads.top, i));
			fnord = 1;
		}
	for (i = 0; i < uploads.top.num; i++)
		if (transfer_calculate(list_index(&uploads.top, i))) {
			transfer_touch(list_index(&uploads.top, i));
			fnord = 1;
		}

	return fnord;
}

int transfer_alive(transfer * t)
{
	return !!t->id;
}

char *transfer_status(transfer * t)
{
	if (t->paused)
		return _("Paused");

	if (!t->id) {
		if (t->transferred == t->filesize)
			return _("Completed");
		if (t->transferred == (unsigned long) -1)
			return _("Cancelled");
		if (t->transferred > t->filesize)
			return _("Overrun");
		return _("Cancelled");
	}

	return "";
}

char *source_status(source * s)
{
	return s->status ? _(s->status) : _("Unknown status");
}

/* NULL in a field act as a wildcard */
source *lookup_source(transfer * t, const char *user, const char *href)
{
	int i;

	for (i = 0; i < t->sourcen.num; i++) {
		source *s = list_index(&t->sourcen, i);
		int match = 1;

		if (user)
			match &= !strcmp(user, s->user);
		if (href)
			match &= !strcmp(href, s->href);
		if (match)
			return s;
	}

	return NULL;
}

void chunk_parse(transfer * t, xml_data buf)
{
	/* <chunk id= source= start= transmit= total= status="Active"/> */
	char *href = xml_get_param(buf, "href");
	unsigned long total;
	source *s = NULL;

	if (!href || !*href) {
		DEBUG("Expected an href attribute.");
		return;
	}
	if (!(total = xml_get_param_int(buf, "total"))) {
		DEBUG("Expected a total attribute.");
		return;
	}
	if (!(s = lookup_source(t, NULL, href))) {
		DEBUG("The source '%s' is not registrered", href);
		return;
	}
	s->transmit = xml_get_param_int(buf, "transmit");
	s->start = xml_get_param_int(buf, "start");
	s->total = total;
	free(s->status);
	s->status = strdup(xml_get_param(buf, "status"));
	TOUCH(s);
}

char *pretty_transfer_original(transfer * t)
{
	int fwidth = max_x - 40;
	char bar[512];
	float ratio;
	unsigned long transferred;

	transferred = t->transferred == (unsigned long) -1 ? 0 : t->transferred;
	ratio = t->filesize ? (float) transferred / t->filesize : 1.0;

	snprintf(bar, sizeof bar, "\\4%*.*s | ", fwidth, fwidth, t->filename);

	if (t->transferred == (unsigned long) -1)
		strncatf(bar, sizeof bar, "???? [ ??%%]");
	else
		strncatf(bar, sizeof bar, "%s [%3d%%]", humanify(t->transferred), (int) (ratio * 100));

	if (t->type == TRANSFER_DOWNLOAD && t->id)
		strncatf(bar, sizeof bar, " %dx", t->sourcen.num);

	strncatf(bar, sizeof bar, " @ %s%s", humanify(bandwidth_shown(t)), _("B/s"));
	strncatf(bar, sizeof bar, " %s", transfer_status(t));

	if (t->type == TRANSFER_DOWNLOAD) {
		if (t->id && t->bandwidth > 0 && !t->paused)
			strncatf(bar, sizeof bar, humanify_time((t->filesize - transferred) / t->bandwidth));
		if (t->specific.d.search_id)
			strncatf(bar, sizeof bar, _(" Searching"));
	}

	strpad(bar + 2, (max_x > sizeof bar ? sizeof bar : max_x) - 4);
	insert_at_ratio(bar + 2, "\\0", ratio, (max_x > sizeof bar ? sizeof bar : max_x) - 4);
	bar[sizeof bar - 1] = '\0';

	return t->pretty = strdup(bar);
}

char *pretty_source_original(transfer * t, source * s)
{
	char bar[512];
	int protolen;

	protolen = strchr(s->href, ':') - s->href;

	snprintf(bar, sizeof bar, "+ %s@%*.*s", s->user, protolen, protolen, s->href);
	strpad(bar, max_x - 40);

	strncatf(bar, sizeof bar, " |");

	if (s->total) {
		strncatf(bar, sizeof bar, " %s [%3d%%]", humanify(s->transmit),
				 (int) (100.0 * s->transmit / s->total));
	}
	strncatf(bar, sizeof bar, " %s", source_status(s));
	if (s->total) {
		strpad(bar + 2, (max_x > sizeof bar ? sizeof bar : max_x) - 4);
		insert_at_ratio(bar, "\\4", (float) s->start / t->filesize,
						(max_x > sizeof bar ? sizeof bar : max_x) - 4);
		insert_at_ratio(bar + 2, "\\0", (float) (s->start + s->transmit) / t->filesize,
						(max_x > sizeof bar ? sizeof bar : max_x) - 4);
	}
	bar[sizeof bar - 1] = '\0';

	return s->pretty = strdup(bar);
}

char *pretty_transfer_jasta(transfer * t)
{
	int prog_w = 35;
	int name_w = max_x - prog_w;
	int prog_x = name_w + 6;
	int slack = 10;

	char bar[512];
	char prog[512];
	float ratio;
	unsigned long transferred;
	char *status = transfer_status(t);

	transferred = t->transferred == (unsigned long) -1 ? 0 : t->transferred;
	ratio = t->filesize ? (float) transferred / t->filesize : 1.0;

	snprintf(bar, sizeof bar, "%c ", t->expanded ? '-' : '+');
	strncatf(bar, name_w + 1, t->filename);
	strpad(bar, sizeof bar - 1);

	if (!strcmp(status, ""))
		snprintf(prog, sizeof prog, "%s/%s %2d%% @ %s%s", humanify(transferred),
				 humanify(t->filesize), (int) (ratio * 100), humanify(bandwidth_shown(t)),
				 _("B/s"));
	else
		snprintf(prog, sizeof prog, "%s", status);

	strncpy(bar + name_w, " | [\\4", 6);
	memcpy(bar + prog_x + (prog_w - strlen(prog) + slack) / 2 - slack, prog, strlen(prog));
	strcpy(bar + prog_x + prog_w - slack, "]");

	insert_at_ratio(bar + prog_x, "\\0", ratio, prog_w - slack);

	bar[sizeof bar - 1] = '\0';
	return t->pretty = strdup(bar);
}

char *pretty_source_jasta(transfer * t, source * s)
{
	int prog_w = 35;
	int name_w = max_x - prog_w;
	int prog_x = name_w;

	char bar[512];
	int protolen;
	char *status = source_status(s);

	protolen = strchr(s->href, ':') - s->href;
	snprintf(bar, sizeof bar, "  - %s@%*.*s", s->user, protolen, protolen, s->href);
	strpad(bar, sizeof bar - 1);
	strpad(bar, 200);

	strncpy(bar + name_w + 1, "| [", 3);
	memcpy(bar + prog_x + prog_w / 2 - strlen(status) / 2 - 1, status, strlen(status));
	strcpy(bar + prog_x + prog_w - 6, "]");

	if (s->total) {
		insert_at_ratio(bar + prog_x + 4, "\\0", (float) (s->start + s->transmit) / t->filesize,
						prog_w - 10);
		insert_at_ratio(bar + prog_x + 4, "\\4", (float) s->start / t->filesize, prog_w - 10);
	}

	bar[sizeof bar - 1] = '\0';
	return s->pretty = strdup(bar);
}



/* How important is the last bandwidth reading? The max is 256 which
 * means that the last reading _is_ the bandwidth.
 */
#define EWMA 64

/* This function tries to estimate the bandwidth a transfer has. To
 * consider is that a new transfer can start in the middle of a file, so
 * we don't want it to say 250MB/s just because it started 250MB into the
 * file. Another thing to consider is that the time between calculations
 * might be zero.
 */
static int transfer_calculate(transfer * t)
{
	unsigned long curspeed = 0;
	unsigned long oldspeed = t->bandwidth;

	if (t->paused) {
		t->bandwidth = 0;
		return !!oldspeed;
	}

	if (t->time == t->last_time)
		return 0;

	if (t->last_trans)
		curspeed = (t->transferred - t->last_trans) / (t->time - t->last_time);

	if (t->bandwidth != (unsigned long) -1)
		t->bandwidth = ((256 - EWMA) * t->bandwidth + EWMA * curspeed) / 256;
	else if (t->transferred != (unsigned long) -1)
		t->bandwidth = curspeed;

	t->last_time = t->time;
	t->last_trans = t->transferred;

	return curspeed != oldspeed;
}

static unsigned long bandwidth_shown(transfer * t)
{
	if (!t->id && t->transferred != (unsigned long) -1) {
		long deltas = t->transferred - t->start_size;

		if (deltas > 0 && t->time > 0)
			return deltas / t->time;
	}
	return t->bandwidth == (unsigned long) -1 ? 0 : t->bandwidth;
}
