/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001-2002 Gran Weinholt <weinholt@linux.nu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: tree.c,v 1.7 2002/06/11 00:54:19 weinholt Exp $
 */
#include "giftcurs.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "tree.h"
#include "screen.h"
#include "parse.h"

static void add_entries(list * flat, list * viflo)
{
	/* recursive helper function */
	int i;

	for (i = 0; i < viflo->num; i++) {
		rendering **r = viflo->entries[i];
		list *(*exp) (void *) = (*r)->expansion;

		list_push(flat, r);
		if (exp && exp(r))
			add_entries(flat, exp(r));
	}
}

void tree_flatten(tree * t)
{
	int i, offset;
	void *selected;

	selected = t->flat.num == 0 || t->flat.sel == -1 ? NULL : list_selected(&t->flat);
	offset = t->flat.sel - t->flat.start;

	list_remove_all(&t->flat);
	add_entries(&t->flat, &t->top);

	if (!selected)
		return;
	/* find the selected. */
	for (i = 0; i < t->flat.num; i++) {
		if (list_index(&t->flat, i) == selected) {
			t->flat.sel = i;
			t->flat.start = i - offset;
			return;
		}
	}
	/* the selected item was lost. */
}

/* Return the parent to selected item, or the item itself if top-level. */
int tree_top_sel(tree * t)
{
	return list_find(&t->top, tree_top_selected(t));
}

void *tree_top_selected(tree * t)
{
	int i = t->flat.sel;

	/* scan upwards until we find the toplevel type */
	while (*(rendering **) list_index(&t->flat, i) != t->top_methods) {
		assert(i);
		i--;
	}
	return list_index(&t->flat, i);
}

/* Find the parent to the child item */
void *tree_find_parent(tree * t, void *item)
{
	int i;

	for (i = 0; i < t->top.num; i++) {
		rendering **r = (rendering **) list_index(&t->top, i);

		list *children = (*r)->expansion ? (*r)->expansion(r) : NULL;

		if (children && list_find(children, item) >= 0)
			return r;
	}
	return NULL;
}

void tree_initialize(tree * t, rendering * r)
{
	list_initialize(&t->flat);
	list_initialize(&t->top);
	t->top_methods = r;
}

void list_destroy_all(list * l)
{
	int i;

	for (i = 0; i < l->num; i++) {
		rendering **r = (rendering **) l->entries[i];

		if (r && (*r)->destroy)
			(*r)->destroy(r);
	}
	list_remove_all(l);
}

void tree_destroy_all(tree * t)
{
	list_remove_all(&t->flat);
	list_destroy_all(&t->top);
}

void list_touch_all(list * l)
{
	int i;

	assert(l);

	for (i = 0; i < l->num; i++) {
		rendering **r = (rendering **) l->entries[i];

		if (r) {
			if ((*r)->touch)
				(*r)->touch(r);

			if ((*r)->expansion) {
				list *expansion = (*r)->expansion(r);

				if (expansion)
					list_touch_all(expansion);
			}
		}
	}
}

void tree_touch_all(tree * t)
{
	list_touch_all(&t->top);
}

void tree_sort(tree * t, SFunc cmp)
{
	void *sel = list_selected(&t->flat);

	qsort(t->top.entries, t->top.num, sizeof(void *), cmp);

	/* find the selected item again */
	t->flat.sel = list_find(&t->flat, sel);
}

void tree_select_top(tree * t, int i)
{
	tree_select_item(t, list_index(&t->top, i));
}

void tree_select_item(tree * t, void *item)
{
	t->flat.sel = list_find(&t->flat, item);
}

void tree_remove_item(tree * t, void *item)
{
	int i = list_find(&t->top, item);

	list_remove_entry(&t->top, i);
	tree_flatten(t);
}

void tree_push(tree * t, void *item)
{
	list_push(&t->top, item);
}

void tree_push_first(tree * t, void *item)
{
	list_push_first(&t->top, item);
}

void list_filter(list * l, int (*keep) (void *))
{
	int i;

	for (i = l->num - 1; i >= 0; i--) {
		rendering **item = l->entries[i];

		if (!keep(item)) {
			if ((*item)->destroy)
				(*item)->destroy(item);
			list_remove_entry(l, i);
		}
	}
}

void tree_filter(tree * t, int (*keep) (void *))
{
	list_filter(&t->top, keep);
	tree_flatten(t);
}
