/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001-2002 Gran Weinholt <weinholt@linux.nu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: ui_help.c,v 1.58 2002/06/23 15:40:04 weinholt Exp $
 */
#include "giftcurs.h"

#include <assert.h>
#include <stdlib.h>

#include "screen.h"
#include "ui_mouse.h"
#include "ui.h"
#include "ui_draw.h"
#include "ui_help.h"
#include "parse.h"

#define __USE_GNU
#include <string.h>

static struct {
	list help;
	int build_width;
} scr = {
LIST_INITIALIZER, -1,};

/* Rules for the help text, so that line split works:
	- Don't use tab
	- Always highlight one word at a time only, or use non-breaking spaces */
const char helptext[] = N_("\
Welcome to giFTcurs!\n\
\n\
Key list:\n\
\\1up\\0/\\1down\\0        select item in list\n\
\\1PgUp\\0/\\1PgDown\\0    scroll list one screen at a time\n\
\\1tab\\0            move to the next input field\n\
\\1F1\\0             this help screen\n\
\\1F2\\0             the main screen\n\
    \\1enter\\0      download the highlighted file\n\
    \\1left\\0/\\1right\\0 change sorting method\n\
    \\1e\\0          expand/collapse search results\n\
\\1F3\\0             the transfer screen\n\
    \\1enter\\0/\\1e\\0    expand/collapse transfer\n\
    \\1j\\0/\\1k\\0        move window splitter\n\
\\1F4\\0             the settings screen\n\
\\1F10\\0            exit the program\n\
\n\
See giFTcurs(1) and giFTcurs.conf(5) for more information.\n\
\n");

/* TRANSLATORS: The authors' names are spellt Gran Weinholt and
   Christian Hggstrm, so if your charset has support for 'a' and 'o'
   with two dots over then please use those characters in their names. */
const char copyright[] = N_("\
Copyright (C) 2001, 2002  Goran Weinholt.\n\
Co-developed by Christian Haggstrom.\n\
Our homepage is at http://giFTcurs.sourceforge.net/.\n\
\n");

const char license[] = "\
This program is free software; you can redistribute it and/or modify\
 it under the terms of version 2 of the GNU General Public License as\
 published by the Free Software Foundation.\n\
\n\
This program is distributed in the hope that it will be useful,\
 but WITHOUT ANY WARRANTY; without even the implied warranty of\
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\
 GNU General Public License for more details.\n\
\n\
You should have received a copy of the GNU General Public License\
 along with this program; if not, write to the Free Software Foundation,\
 Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.";

static void help_screen_resize(void);
static void help_screen_update(void);

#ifdef MOUSE
/* Mouse callback */
static void mouse_helpclick(int rx, int ry, void *data);
#endif

void help_screen_init(void)
{
	list_initialize(&scr.help);
	help_screen_resize();
}

static void help_screen_resize(void)
{
	/* TODO: don't count formatting codes as chars. */
	char *text, *fnord;
	int width = max_x - 4;

	fnord = sprintf_alloc("%s%s%s", _(helptext), _(copyright), license);

	list_free_entries(&scr.help);
	text = fnord;
	for (;;) {
		char *pos = strchr(text, '\n');

		if (pos && pos <= text + width) {
			list_push(&scr.help, strndup(text, pos - text));
			text = pos + 1;
			continue;
		}
		pos = strchr(text, '\0');
		if (pos && pos <= text + width) {
			list_push(&scr.help, strdup(text));
			break;
		}
		for (pos = text + width; pos > text; pos--) {
			if (*pos == ' ') {
				list_push(&scr.help, strndup(text, pos - text));
				text = pos + 1;
				break;			/* continue the outer loop */
			}
		}
		if (text == pos + 1)
			continue;
		list_push(&scr.help, strndup(text, width));
		text += width;
	}
	scr.build_width = max_x;
	free(fnord);
}

void help_screen_destroy(void)
{
	list_free_entries(&scr.help);
}

static void help_screen_draw(void)
{
	clrscr();
	curs_set(0);
	leaveok(stdscr, TRUE);
	draw_box(0, 0, max_x, max_y - 1 - show_buttonbar, _("Help"),
			 COLOR_PAIR(COLOR_HELP_BOX) | A_BOLD);
	mouse_clear();
	mouse_register(0, 0, max_x, max_y - 1 - show_buttonbar, BUTTON1_PRESSED, mouse_helpclick, NULL);
	help_screen_update();
}

static void help_screen_update(void)
{
	if (max_x != scr.build_width)
		help_screen_resize();
	list_check_values(&scr.help, max_y - 3 - show_buttonbar);
	scr.help.start = scr.help.sel;
	draw_list(2, 1, max_x - 4, max_y - 3 - show_buttonbar, 1 != 1, &scr.help);
	refresh();
}

static int help_screen_handler(int key)
{
	ui_list_handler(&scr.help, key, max_y - 3);
	help_screen_update();

	return 0;
}

#ifdef MOUSE
static void mouse_helpclick(int rx, int ry, void *data)
{
	if (active_screen != &help_screen_methods)
		return;
	if (ry < max_y / 2)
		scr.help.sel -= 3;
	else
		scr.help.sel += 3;
	list_check_values(&scr.help, max_y - 3 - show_buttonbar);
	help_screen_update();
}
#endif

ui_methods help_screen_methods = {
	help_screen_draw, help_screen_handler
};
