/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001-2002 Gran Weinholt <weinholt@linux.nu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: ui_main.c,v 1.236 2002/07/05 14:50:31 weinholt Exp $
 */
#include "giftcurs.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "gift.h"
#include "stats.h"
#include "screen.h"
#include "ui_mouse.h"
#include "ui.h"
#include "ui_draw.h"
#include "ui_main.h"
#include "get.h"
#include "parse.h"

#define HITS (((query*)list_selected(&queries))->hits)

#define INPUT_MAX		200

enum {
	FIELD_SEARCH_TERM,
	FIELD_SEARCH_REALM,
	FIELD_SEARCH_BUTTON,
	FIELD_STOP_BUTTON,
	FIELD_RESULT_LIST,
	FIELD_MAX,
	FIELD_STATUSBAR,
};

static struct {
	int active_field;
	char search_term[INPUT_MAX];
	unsigned int search_pos;
	int curs_x, curs_y;
	list search_realm;
	char *slash_search_term;
	int slash_search_pos;
	char *csize_term, *cspeed_term;
	int csize_pos, cspeed_pos;
	unsigned long maxsize, minsize;
	int maxspeed, minspeed;
	int pretty_width;
} scr = {
FIELD_SEARCH_TERM, "", 0, 0, 0, LIST_INITIALIZER, NULL, 0, NULL, NULL, 0, 0, 0, 0, 0, -1,};

static void main_screen_draw(void);
static void main_screen_update_stats(void);
static void main_screen_update_results(void);
static void main_screen_update_info(void);
static void main_screen_update_searchbox(void);

#ifdef MOUSE
/* Mouse event callbacks */
static void mouse_infoclick(int rx, int ry, void *data);
static void mouse_resultclick(int rx, int ry, void *data);
static void mouse_buttonclick(int rx, int ry, void *data);
static void mouse_inputclick(int rx, int ry, void *data);
static void mouse_wheel(int rx, int ry, void *data);
#endif

/* The stat box size is fixed. Other boxes may vary in size. */
#define STAT_BOX_W		26
#define STAT_BOX_H		6
#define STAT_HEADLEN	12
#define STAT_COL1		(max_x - STAT_BOX_W)
#define STAT_COL2		(STAT_COL1 + STAT_HEADLEN + 2)
#define SEARCH_BOX_W	(STAT_COL1 - 2)
#define RESULT_H		(max_y - 3 - STAT_BOX_H - show_buttonbar)

void main_screen_update_stats_if_focus(void)
{
	if (active_screen == &main_screen_methods) {
		main_screen_update_stats();
		refresh();
	}
}

void main_screen_init(void)
{
	unsigned int i;
	static const char *realms[] = {
		N_("everything"), N_("audio"), N_("video"),
		N_("images"), N_("text documents"), N_("software"),
		N_("user"),
		N_("hash"),
	};

	list_initialize(&scr.search_realm);

	for (i = 0; i < buflen(realms); i++)
		list_push(&scr.search_realm, strdup(_(realms[i])));
	scr.slash_search_term = NULL;
	scr.active_field = 0;
}

void main_screen_destroy(void)
{
	gift_stat_stop();
	free(scr.slash_search_term);
	list_free_entries(&scr.search_realm);
}

void incoming_search_item(void)
{
	if (active_screen == &main_screen_methods) {
		tree_flatten(&HITS);
		if (scr.active_field != FIELD_RESULT_LIST)
			main_screen_update_searchbox();
		main_screen_update_results();
		refresh();
	}
}

static void prepare_search(void)
{
	int i;
	query *q;

	/* trim it and make cursor inside */
	trim(scr.search_term);
	if (scr.search_pos > strlen(scr.search_term))
		scr.search_pos = strlen(scr.search_term);
	/* add query to history if it's not already there */
	/* skip the first dummy entry */
	for (i = 1; i < queries.num; i++) {
		q = list_index(&queries, i);

		if (!strcmp(scr.search_term, q->search_term) && scr.search_realm.sel == q->realm) {
			queries.sel = i;
			gift_query_stop(q);
			gift_query_clear(q);
			return;
		}
	}

	/* create a new query structure and append to list */
	q = calloc(1, sizeof(query));
	q->search_term = strdup(scr.search_term);
	q->realm = scr.search_realm.sel;
	q->id = NULL;
	tree_initialize(&q->hits, &hit_methods);
	i--;
	list_insert(&queries, q, i);
	queries.sel = i;
}

void start_search(char *query, const char *constraint)
{
	static const char *names[] = {
		NULL,
		"realm=audio",
		"realm=video",
		"realm=image",
		"realm=text",
		"realm=application",
		"type=user",
		"type=hash",
	};
	char buf[512];

	if (query)
		strncpy(scr.search_term, query, INPUT_MAX);
	if (constraint) {
		int i = lookup(constraint, names, scr.search_realm.num);

		if (i != -1)
			scr.search_realm.sel = i;
	} else
		constraint = names[scr.search_realm.sel];

	prepare_search();

	snprintf(buf, sizeof buf, "query=\"%s\"", scr.search_term);

	if (constraint)
		strncatf(buf, sizeof buf, " %s", constraint);

	/*
	   if (filter_string)
	   strncatf(buf, sizeof buf, " exclude=\"%s\"", filter_string);
	 */

	/* these should be replaced with some more intelligent ones */
	if (scr.csize_term && *scr.csize_term)
		strncatf(buf, sizeof buf, " csize=\"%s\"", scr.csize_term);

	if (scr.cspeed_term && *scr.cspeed_term)
		strncatf(buf, sizeof buf, " ckbps=\"%s\"", scr.cspeed_term);

	if (gift_query_start(buf) < 0)
		message(_("Couldn't start search! Check the log for errors."));
	else
		message(_("Search in progress..."));
}

void slash_find_next(int dir)
{
	tree *t = &HITS;
	int prev = tree_top_sel(t);
	int i = prev;
	char *pattern = scr.slash_search_term;

	if (!pattern || !pattern[0] || !pattern[1]) {
		message(_("No previous search pattern."));
		return;
	}
	if (tree_isempty(t)) {
		message(_("No hits to search on."));
		return;
	}

	if (pattern[0] == '?')
		dir = -dir;

	pattern++;

	for (;;) {
		hit *h;

		i += dir;
		if (i >= t->top.num)
			i = 0;
		if (i < 0)
			i = t->top.num - 1;
		if (i == prev) {
			message(_("Pattern not found: '%s'."), pattern);
			return;
		}
		h = list_index(&t->top, i);
		if (stristr(h->filename, pattern)) {
			tree_select_top(&HITS, i);
			message(scr.slash_search_term);
			return;
		}
	}
}

/* returns 0 - can't, 1 - can clear, 2 - can stop */
static int can_stop_search(void)
{
	query *q;

	if (!queries.num)
		return 0;
	q = list_selected(&queries);
	if (q->id)
		return 2;
	return !tree_isempty(&q->hits);
}

static int main_screen_handler(int key)
{
	int ret = 0;

	if (key == '\t' || key == KEY_BTAB) {
		int direction = key == KEY_BTAB ? -1 : 1;

		for (;;) {
			scr.active_field += direction;
			if (scr.active_field >= FIELD_MAX)
				scr.active_field -= FIELD_MAX;
			if (scr.active_field < 0)
				scr.active_field += FIELD_MAX;
			if (scr.active_field == FIELD_STOP_BUTTON && !can_stop_search())
				continue;
			if (scr.active_field == FIELD_RESULT_LIST && (!queries.num || tree_isempty(&HITS)))
				continue;
			break;
		}

		if (scr.active_field == FIELD_RESULT_LIST && HITS.flat.sel == -1)
			HITS.flat.sel++;

		ui_draw();
		refresh();
		return 0;
	}

	switch (scr.active_field) {
	case FIELD_SEARCH_TERM:
		if (!ui_input_handler(scr.search_term, &scr.search_pos, key, INPUT_MAX - 1))
			if (queries.num)
				if (ui_list_handler(&queries, key, 5)) {
					query *q = list_selected(&queries);

					scr.search_pos = INPUT_MAX;	/* place cursor at the end */
					strcpy(scr.search_term, q->search_term);
					scr.search_realm.sel = q->realm;
					/* background searches are not sorted - sort it now */
					change_sort_method(0);
				}
		/* Check values */
		ui_input_handler(scr.search_term, &scr.search_pos, 0, INPUT_MAX - 1);
		if (key == KEY_ENTER && *scr.search_term) {
			scr.active_field = FIELD_SEARCH_BUTTON;
			start_search(NULL, NULL);
			main_screen_update_searchbox();
			main_screen_update_results();
			refresh();
			return 0;
		}

		main_screen_update_searchbox();
		refresh();
		break;
	case FIELD_SEARCH_REALM:
		if (ui_list_handler(&scr.search_realm, key, 1)) {
			list_check_values(&scr.search_realm, 1);

			main_screen_update_searchbox();
			refresh();
		} else if (key == KEY_ENTER && *scr.search_term) {
			scr.active_field = FIELD_SEARCH_BUTTON;
			start_search(NULL, NULL);
			main_screen_update_searchbox();
			main_screen_update_results();
			refresh();
			return 0;
		}
		break;
	case FIELD_SEARCH_BUTTON:
		if (PRESSED(key)) {
			start_search(NULL, NULL);
			main_screen_update_searchbox();
			main_screen_update_results();
			refresh();
		}
		break;
	case FIELD_STOP_BUTTON:
		if (PRESSED(key)) {
			query *q;

			if (!queries.num)
				break;
			q = list_selected(&queries);

			if (q->id) {
				gift_query_stop(q);
				message(_("Search stopped."));
			} else if (!tree_isempty(&q->hits)) {
				gift_query_clear(q);
				message(_("Search cleared."));
				main_screen_update_results();
			} else {
				message(_("A magical search appears, but it is cancelled too soon."));
			}
			main_screen_update_searchbox();
			refresh();
		}
		break;
	case FIELD_RESULT_LIST:
		if (PRESSED(key) || key == 'd' || key == 'D') {
			int i = tree_top_sel(&HITS);

			message(_("Downloading..."));
			download_hit(list_index(&HITS.top, i));
			if (i < HITS.top.num - 1)
				i++;
			tree_select_top(&HITS, i);
			main_screen_update_results();
			refresh();
		} else if (key == KEY_LEFT || key == KEY_RIGHT) {
			const char *method = change_sort_method(key == KEY_RIGHT ? 1 : -1);

			message(_("Sorting order changed to %s."), method);
		} else if (ui_list_handler(&HITS.flat, key, RESULT_H)) {
			main_screen_update_results();
			refresh();
		} else if (key == '/' || key == '?') {
			scr.active_field = FIELD_STATUSBAR;
			if (!scr.slash_search_term)
				scr.slash_search_term = calloc(INPUT_MAX, 1);
			scr.slash_search_term[0] = key;
			scr.slash_search_pos = strlen(scr.slash_search_term);
			main_screen_update_results();
			refresh();
		} else if (key == 'n' || key == 'N') {
			slash_find_next(key == 'n' ? 1 : -1);
			main_screen_update_results();
			refresh();
		} else if (key == 'E' || key == 'e' || key == '+' || key == '-') {
			/* (un) expand this item. */
			hit *h = tree_top_selected(&HITS);

			if (key == '+' || key == '-')
				h->expanded = key == '+';
			else
				h->expanded = !h->expanded;
			TOUCH(h);
			tree_select_item(&HITS, h);
			tree_flatten(&HITS);
			main_screen_update_results();
			refresh();
		} else if (key == 'U' || key == 'u') {
			hit *h = list_selected(&HITS.flat);

			if (!IS_A(h, subhit))
				break;
			start_search(((subhit *) h)->user, "type=user");
		}
		break;
	case FIELD_STATUSBAR:
		if (!ui_input_handler(scr.slash_search_term, &scr.slash_search_pos, key, INPUT_MAX - 1)) {
			scr.active_field = FIELD_RESULT_LIST;
			if (PRESSED(key))
				slash_find_next(1);
		}
		main_screen_update_results();
		refresh();
		break;
	}

	return ret;
}

static void main_screen_draw(void)
{
	int y = 1, attr;

	if (active_screen != &main_screen_methods)
		return;

	/* invalidate all pretty-strings if screen width has changed. */
	if (scr.pretty_width != max_x) {
		tree_touch_all(&HITS);
		scr.pretty_width = max_x;
	}

	if (scr.active_field > FIELD_MAX)
		scr.active_field = FIELD_RESULT_LIST;

	curs_set(0);
	leaveok(stdscr, TRUE);
	mouse_clear();

	if (scr.active_field != FIELD_RESULT_LIST) {
		/* Searchbox */
		attr = COLOR_PAIR(COLOR_SEARCH_BOX) | A_BOLD;
		draw_box(0, 0, SEARCH_BOX_W, STAT_BOX_H, _("Search"), attr);
		main_screen_update_searchbox();
	} else {
		attr = COLOR_PAIR(COLOR_INFO_BOX) | A_BOLD;
		draw_box(0, 0, SEARCH_BOX_W, STAT_BOX_H, _("Extra Information"), attr);
		mouse_register(0, 0, SEARCH_BOX_W, STAT_BOX_H, BUTTON1_PRESSED, mouse_infoclick, NULL);
	}

	/* Resultbox */
	tree_flatten(&HITS);
	draw_box(0, STAT_BOX_H, max_x, RESULT_H + 2, _("Search Results"),
			 COLOR(COLOR_RESULT_BOX, FIELD_RESULT_LIST));
	if (scr.active_field == FIELD_RESULT_LIST)
		scr.curs_x = scr.curs_y = 0;
	main_screen_update_results();
	mouse_register(0, STAT_BOX_H, max_x, RESULT_H + 2, BUTTON1_PRESSED, mouse_resultclick, NULL);
	mouse_register(0, 0, max_x, max_y, BUTTON_UP, mouse_wheel, (void *) 0);
	mouse_register(0, 0, max_x, max_y, BUTTON_DOWN, mouse_wheel, (void *) 1);

	/* Statbox */
	attr = COLOR_PAIR(COLOR_STAT_BOX) | A_BOLD;
	draw_box(max_x - STAT_BOX_W - 2, 0, STAT_BOX_W + 2, STAT_BOX_H, _("Statistics"), attr);
	attrset(COLOR_PAIR(COLOR_HEADER) | A_BOLD);
	draw_header(STAT_COL1, y++, _("Users online"));
	draw_header(STAT_COL1, y++, _("Total shared"));
	draw_header(STAT_COL1, y++, _("Local shares"));
	draw_header(STAT_COL1, y++, _("giFT plugins"));
	main_screen_update_stats();

	move(scr.curs_y, scr.curs_x);

	/* Haven't found a better place for this yet... */
	gift_stat_update(main_screen_update_stats_if_focus);
}

static void draw_input_special(int x, int y, int w, char *header, char *value, int field, int pos)
{
	int xx, offset = 0, headlen = strlen(header);

	if (scr.active_field == field && pos > w - headlen - 5)
		offset = pos - (w - headlen - 5);

	xx = draw_input(x, y, w, header, value + offset, COLOR(COLOR_SEARCH_BOX, field));

	if (scr.active_field == field) {
		scr.curs_x = xx;
		if (pos >= 0)
			scr.curs_x = x + pos - offset + headlen + 3;
		getyx(stdscr, scr.curs_y, xx);
		curs_set(1);
		leaveok(stdscr, FALSE);
	}
}

static void main_screen_update_searchbox(void)
{
	int len;

	curs_set(0);
	leaveok(stdscr, TRUE);

	draw_input_special(2, 1, SEARCH_BOX_W - 4, _("Query"),
					   scr.search_term, FIELD_SEARCH_TERM, scr.search_pos);

	draw_input_special(2, 2, SEARCH_BOX_W - 4, _("Realm"),
					   list_selected(&scr.search_realm), FIELD_SEARCH_REALM, -1);

	len = strlen(_("Query"));
	mouse_register(len + 4, 1, max_x - STAT_BOX_W - len - 8, 1, BUTTON1_PRESSED, mouse_inputclick,
				   (void *) FIELD_SEARCH_TERM);
	len = strlen(_("Realm"));
	mouse_register(len + 4, 2, max_x - STAT_BOX_W - len - 8, 1, BUTTON1_PRESSED, mouse_inputclick,
				   (void *) FIELD_SEARCH_REALM);

	/* since we have resizable buttons now, we need to clear the area first */
	mvhline(STAT_BOX_H - 2, 1, ' ', SEARCH_BOX_W - 2);

	draw_button(SEARCH_BOX_W / 4 + 1, STAT_BOX_H - 2, _("Start search"),
				COLOR_BUTTON(COLOR_SEARCH_BOX, FIELD_SEARCH_BUTTON, 1), mouse_buttonclick,
				(void *) FIELD_SEARCH_BUTTON);

	draw_button(3 * SEARCH_BOX_W / 4, STAT_BOX_H - 2,
				can_stop_search() > 1 ? _("Stop search") : _("Clear search"),
				COLOR_BUTTON(COLOR_SEARCH_BOX, FIELD_STOP_BUTTON, can_stop_search()),
				mouse_buttonclick, (void *) FIELD_STOP_BUTTON);

	move(scr.curs_y, scr.curs_x);
}

static void main_screen_update_info(void)
{
	hit *info;
	int y = 1, x = 0, slack;
	char *href;
	int i, j;
	subhit *sh;

	clear_area(1, 1, SEARCH_BOX_W - 2, STAT_BOX_H - 2);

	if (HITS.flat.sel < 0) {
		mvaddstr(y, 2, _("Nothing selected."));
		return;
	}
	info = tree_top_selected(&HITS);
	draw_header(2, y++, _("Size"));
	getyx(stdscr, slack, x);
	draw_printf(SEARCH_BOX_W - x - 2, ngettext("%lu byte", "%lu bytes", info->filesize),
				info->filesize);
	getyx(stdscr, slack, x);

	draw_header(2, y++, _("Net"));
	sh = list_index(&info->sources, 0);
	href = sh->href;
	i = strchr(href, ':') - href;
	if (i > SEARCH_BOX_W - 4)
		i = SEARCH_BOX_W - 4;
	addnstr(href, i);

	draw_header(2, y++, _("User"));
	getyx(stdscr, slack, x);
	addnstr(sh->user, SEARCH_BOX_W - x);
	j = strlen(sh->user);
	for (i = 1; i < info->sources.num && j < SEARCH_BOX_W - x - 2; i++) {
		sh = list_index(&info->sources, i);
		j += draw_printf(SEARCH_BOX_W - x - 2 - j, ", %s", sh->user);
	}

	if (info->directory) {
		draw_header(2, y++, _("Path"));
		getyx(stdscr, slack, x);
		addnstr(info->directory, SEARCH_BOX_W - x - 2);
	}
}

static void draw_critic_stat(int x, int y, int w, unsigned int val)
{
	if (val) {
		attrset(COLOR_PAIR(COLOR_STAT_DATA));
		move(y, x);
		draw_printfmt(w, "\\s%d", val);
	} else {
		attrset(COLOR_PAIR(COLOR_STAT_DATA_BAD));
		mvaddch(y, x, '0');
	}
}

static void main_screen_update_stats(void)
{
	int y = 1;
	int width = max_x - STAT_COL2 - 1;

	clear_area(STAT_COL2, y, width, STAT_BOX_H - 2);

	draw_critic_stat(STAT_COL2, y++, width, stats.users);

	attrset(COLOR_PAIR(COLOR_STAT_DATA));
	move(y++, STAT_COL2);
	draw_printfmt(width, _("\\s%sB\\S/\\s%s"), humanify_scale(stats.gigs, 3),
				  humanify(stats.files));

	move(y++, STAT_COL2);
	draw_printfmt(width, _("\\s%sB\\S/\\s%s"), stats.own_gigs || !stats.own_files
				  ? humanify_scale(stats.own_gigs, 3) : " <1G", humanify(stats.own_files));

	draw_critic_stat(STAT_COL2, y++, width, stats.protos_up);
	attrset(COLOR_PAIR(COLOR_STAT_DATA));
	draw_printfmt(width - 2, "\\S/\\s%i", stats.protos);

	move(scr.curs_y, scr.curs_x);
}

/* This function creates the ->pretty string for search results. */
static char *gift_hit_present(hit * foo)
{
	char tmp[512];
	int c;
	int avail = 0, i;

	if (foo->pretty)
		return foo->pretty;

	for (i = 0; i < foo->sources.num; i++)
		avail += !!((subhit *) list_index(&foo->sources, i))->availability;

	if (foo->downloading)
		c = 0;
	else if (avail >= 2)
		c = 2;
	else if (avail == 0)
		c = 3;
	else
		c = 1;

	snprintf(tmp, sizeof tmp, "%c \\%i%2i/%-2i\\%i %s\\0 %s", foo->expanded ? '-' : '+', c, avail,
			 foo->sources.num, !foo->downloading, humanify(foo->filesize), foo->filename);
	return (foo->pretty = strdup(tmp));
}

/* TODO: Make this colorful */
static char *gift_subhit_present(subhit * foo)
{
	char tmp[512];
	char *filename;
	int protolen;
	int c;

	if (foo->pretty)
		return foo->pretty;

	if (foo->availability == 0)
		c = 3;
	else if (foo->availability > 1)
		c = 2;
	else
		c = 0;
	protolen = strchr(foo->href, ':') - foo->href;
	filename = extract_filename_alloc(foo->href);
	snprintf(tmp, sizeof tmp, " -\\%i % 3i \\0    %16s\\1@\\0%*.*s: %s", c, foo->availability,
			 foo->user, protolen, protolen, foo->href, filename);
	free(filename);
	return (foo->pretty = strdup(tmp));
}

static list *hit_expansion(hit * h)
{
	return h->expanded ? &h->sources : NULL;
}

static void hit_touch(hit * h)
{
	TOUCH(h);
}

static void subhit_touch(subhit * h)
{
	TOUCH(h);
}

rendering hit_methods = {
	(char *(*)(void *)) gift_hit_present,
	(list * (*)(void *)) hit_expansion,
	(void (*)(void *)) hit_destroy,
	(void (*)(void *)) hit_touch,
};
rendering subhit_methods = {
	(char *(*)(void *)) gift_subhit_present,
	NULL,
	(void (*)(void *)) subhit_destroy,
	(void (*)(void *)) subhit_touch,
};

static void main_screen_update_results(void)
{
	char tmp[15];
	int i, h = RESULT_H;
	query *q;
	tree *t;

	if (!queries.num)
		return;

	q = list_selected(&queries);
	t = &q->hits;

	/* Update resultbox, keep cursor invisible if not activated */
	if (t->flat.sel != -1)
		list_check_values(&t->flat, h);

	/* quick hack to show the cursor even if browsing not begun */
	i = t->flat.sel;
	if (i == -1)
		t->flat.sel++;
	draw_list_pretty(2, STAT_BOX_H + 1, max_x - 4, h,
					 scr.active_field >= FIELD_RESULT_LIST, &t->flat);
	t->flat.sel = i;

	move(STAT_BOX_H + RESULT_H + 1, 3);
	attrset(COLOR_PAIR(COLOR_RESULT_BOX) | (scr.active_field >= FIELD_RESULT_LIST ? A_BOLD : 0));
	/* find the selected item. FIXME: this is too expensive! */
	if (i != -1 && !tree_isempty(t))
		i = tree_top_sel(t);
	snprintf(tmp, sizeof tmp, " %i/%i ", i + 1, t->top.num);
	addstr(tmp);
	for (i = strlen(tmp); i < sizeof tmp; i++)
		addch(ACS_HLINE);

	if (scr.active_field == FIELD_STATUSBAR) {
		show_status(scr.slash_search_term);
		scr.curs_y = max_y - 1 - show_buttonbar;
		scr.curs_x = scr.slash_search_pos;
		curs_set(1);
		leaveok(stdscr, FALSE);
	} else if (scr.active_field == FIELD_RESULT_LIST) {
		curs_set(0);
		leaveok(stdscr, TRUE);
		main_screen_update_info();
	}

	move(scr.curs_y, scr.curs_x);
	q->dirty = 0;
}

#ifdef MOUSE
static void mouse_infoclick(int rx, int ry, void *data)
{
	if (active_screen != &main_screen_methods)
		return;
	scr.active_field = FIELD_SEARCH_TERM;
	ui_draw();
	refresh();
}

static void mouse_resultclick(int rx, int ry, void *data)
{
	static int lastrx = -1;
	static int lastry = -1;
	static tick_t lastuptime = 0;

	if (active_screen != &main_screen_methods || !queries.num || tree_isempty(&HITS))
		return;

	if (rx == lastrx && ry == lastry && (uptime() - lastuptime) < 20) {
		download_hit(list_index(&HITS.top, tree_top_sel(&HITS)));
		lastuptime = 0;
		return;
	}
	lastrx = rx;
	lastry = ry;
	lastuptime = uptime();

	if (scr.active_field != FIELD_RESULT_LIST) {
		scr.active_field = FIELD_RESULT_LIST;
		ui_draw();
	}
	if (ry == 0)
		HITS.flat.sel -= RESULT_H;
	else if (ry == RESULT_H + 1)
		HITS.flat.sel += RESULT_H;
	else
		HITS.flat.sel = HITS.flat.start + ry - 1;

	/* Expand/collapse a search result. */
	if (rx == 2)
		ui_handler('e');

	list_check_values(&HITS.flat, RESULT_H);
	main_screen_update_results();
	main_screen_update_info();
	refresh();
}

static void mouse_buttonclick(int rx, int ry, void *data)
{
	int field = (int) data;

	if (active_screen != &main_screen_methods)
		return;

	scr.active_field = field;
	main_screen_handler(KEY_ENTER);
	ui_draw();
	refresh();
}

static void mouse_inputclick(int rx, int ry, void *data)
{
	int field = (int) data;

	if (active_screen != &main_screen_methods)
		return;

	if (field == FIELD_SEARCH_TERM) {
		scr.search_pos = rx - 1;
		ui_input_handler(scr.search_term, &scr.search_pos, 0, INPUT_MAX - 1);
	}

	scr.active_field = field;
	ui_draw();
	refresh();
}

static void mouse_wheel(int rx, int ry, void *data)
{
	int dir = (int) data;

	if (active_screen != &main_screen_methods || tree_isempty(&HITS))
		return;

	/* TODO: support scrolling in more fields? */
	if (scr.active_field != FIELD_RESULT_LIST)
		return;

	if (dir)
		HITS.flat.sel += (RESULT_H / 2);
	else
		HITS.flat.sel -= (RESULT_H / 2);

	list_check_values(&HITS.flat, RESULT_H);
	main_screen_update_results();
	main_screen_update_info();
	refresh();
}
#endif

ui_methods main_screen_methods = {
	main_screen_draw, main_screen_handler
};
