/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001-2002 Gran Weinholt <weinholt@linux.nu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: ui_transfer.c,v 1.180 2002/06/30 01:51:06 weinholt Exp $
 */
#include "giftcurs.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#if HAVE_STATFS
# include <sys/types.h>
# include <sys/param.h>
# if HAVE_SYS_MOUNT_H
#  include <sys/mount.h>
# endif
# if HAVE_SYS_VFS_H
#  include <sys/vfs.h>
# endif
# if HAVE_SYS_STATVFS_H
#  include <sys/statvfs.h>
#  define statfs statvfs
# endif
#endif

#include "screen.h"
#include "gift.h"
#include "ui_mouse.h"
#include "ui.h"
#include "ui_draw.h"
#include "ui_main.h"
#include "ui_transfer.h"
#include "transfer.h"
#include "get.h"
#include "upload.h"
#include "parse.h"
#include "poll.h"
#include "settings.h"

enum transfer_screen_fields {
	FIELD_DOWNLOADS,
	FIELD_UPLOADS,
	FIELD_MAX
};

enum {
	STYLE_ORIGINAL,
	STYLE_JASTA
};

/* *INDENT-OFF* */
static struct {
	int style_int;
	char *style_char;
} styles[] = {
	{STYLE_ORIGINAL, "original"},
	{STYLE_JASTA, "jasta"},
	{0, NULL}
};
/* *INDENT-ON* */

static struct {
	int active_field;
} scr = {
FIELD_DOWNLOADS};

static void transfer_screen_uploads_update(void);
static void transfer_screen_downloads_update(void);
static void transfer_screen_draw(void);
static int transfer_screen_timer(void *data);

#ifdef MOUSE
/* Mouse callbacks */
static void mouse_listclick(int rx, int ry, void *data);
static void mouse_wheel(int rx, int ry, void *data);
#endif

static int UPLOAD_H = 10;
static int transfer_style = STYLE_ORIGINAL;

#define DOWNLOAD_H	(max_y - UPLOAD_H - 2 - show_buttonbar)
#define HELP_PAD	2

void transfer_screen_init(void)
{
	char *style;
	int i;

	UPLOAD_H = atoi(get_config("set", "upload-height", "10"));
	style = get_config("set", "transfer-style", "jasta");
	for (i = 0; styles[i].style_char; i++)
		if (!strcmp(styles[i].style_char, style)) {
			transfer_style = styles[i].style_int;
			break;
		}
	if (!styles[i].style_char)
		transfer_style = STYLE_JASTA;

	poll_add_timer(SECS(1), transfer_screen_timer, NULL);
}

void transfer_screen_destroy(void)
{
	tree_destroy_all(&downloads);
	tree_destroy_all(&uploads);
	poll_del_timer(transfer_screen_timer);
}

static int transfer_screen_handler(int key)
{
	transfer *t;
	tree *active_tree;
	int height;
	void (*update_func) (void);

	switch (key) {
	case '\t':
	case KEY_BTAB:
		/* when we have more than two fields, take care of shift-tab here */
		scr.active_field++;
		if (scr.active_field >= FIELD_MAX)
			scr.active_field -= FIELD_MAX;
		ui_draw();
		refresh();
		return 0;
	case 'C':
	case 'c':
		tree_filter(&downloads, (int (*)(void *)) transfer_alive);
		tree_filter(&uploads, (int (*)(void *)) transfer_alive);
		ui_draw();
		refresh();
		return 0;
	case 'K':
	case 'k':
		if (DOWNLOAD_H > 3) {
			UPLOAD_H++;
			set_config_int("set", "upload-height", UPLOAD_H, 1);
			ui_draw();
			refresh();
		}
		return 0;
	case 'J':
	case 'j':
		if (UPLOAD_H > 3) {
			UPLOAD_H--;
			set_config_int("set", "upload-height", UPLOAD_H, 1);
			ui_draw();
			refresh();
		}
		return 0;
	}

	/* This is getting clearer, but still ugly... */
	if (scr.active_field == FIELD_DOWNLOADS) {
		active_tree = &downloads;
		height = DOWNLOAD_H;
		update_func = transfer_screen_downloads_update;
	} else {
		active_tree = &uploads;
		height = UPLOAD_H;
		update_func = transfer_screen_uploads_update;
	}
	if (tree_isempty(active_tree))
		return 0;
	list_check_values(&active_tree->flat, height - 3);

	t = list_selected(&active_tree->flat);

	switch (key) {
	case 'P':
	case 'p':
		if (!IS_A(t, transfer))
			break;
		if (transfer_alive(t))
			transfer_suspend(t);
		update_func();
		refresh();
		break;
	case 'E':
	case 'e':
	case '-':
	case '+':
	case KEY_ENTER:
		if (IS_A(t, source)) {
			int i = tree_top_sel(active_tree);

			t = list_index(&active_tree->top, i);
			tree_select_top(active_tree, i);
		}
		if (key == '+' || key == '-')
			t->expanded = key == '+';
		else
			t->expanded = !t->expanded;
		TOUCH(t);
		update_func();
		refresh();
		break;
	case 'S':
	case 's':
		/* Issue a new search if one is not already running. Also
		   don't search if the download is finished. */
		if (!IS_A(t, transfer) || t->type != TRANSFER_DOWNLOAD)
			break;
		if (t->transferred != t->filesize) {
			download_search(t);
			transfer_screen_downloads_update();
			refresh();
		}
		break;
	case 'T':
	case 't':
		if (IS_A(t, source)) {
			source *s = (source *) t;

			t = tree_top_selected(active_tree);
			download_delsource(t, s->href);
			gift_response(t->id, "transfer user=\"%s\" delsource=\"%s\"", s->user, s->href);
			active_tree->flat.sel--;
		} else {
			if (transfer_alive(t))
				transfer_cancel(t);
			else {
				if (t->type == TRANSFER_DOWNLOAD)
					download_forget(t);
				else
					upload_forget(t);
			}
		}
		update_func();
		refresh();
		break;
	case 'U':
	case 'u':
		if (!IS_A(t, source))
			break;
		start_search(((source *) t)->user, "type=user");
		break;
	default:
		if (ui_list_handler(&active_tree->flat, key, height - 2)) {
			update_func();
			refresh();
		}
	}
	return 0;
}

#if HAVE_STATFS
#define STATFS_INTERVAL SECS(5)
static unsigned long disk_free_kb(void)
{
	/* Cache values so we don't need to call statfs so often */
	/* However, we call gettimeofday() more often ... */
	static unsigned long cached_free_kb;
	static tick_t cache_expires;

	tick_t now = uptime();

	if (now > cache_expires) {
		struct statfs statbuf;

		cache_expires = now + STATFS_INTERVAL;

		if (statfs(download_dir, &statbuf) == -1) {
			ERROR("statfs");
			cached_free_kb = -1;
			return -1;
		}
		cached_free_kb = (statbuf.f_bsize / 256) * statbuf.f_bavail / 4;
	}
	return cached_free_kb;
}
#endif

static void transfer_screen_draw(void)
{
    /* *INDENT-OFF* */
    static struct {
        const int key; 
        const char *desc;
    } help[] = {
        {'S', N_("find more sources")},
        {'T', N_("kill transfer")},
        {'C', N_("clear finished")},
        {'P', N_("(un)pause")},
        {'\0', NULL}
    };
    /* *INDENT-ON* */
	static int len = -1;
	static int formatted_xdim = -1;
	int x1, w1, i;

	clrscr();
	curs_set(0);
	leaveok(stdscr, TRUE);

	/* Check that splitter has a sane value */
	if (DOWNLOAD_H < 3)
		UPLOAD_H = max_y - 5;

	draw_box(0, 0, max_x, DOWNLOAD_H, _("Downloads"), COLOR(COLOR_DOWNLOAD_BOX, FIELD_DOWNLOADS));
	draw_box(0, DOWNLOAD_H + 1, max_x, UPLOAD_H, _("Uploads"),
			 COLOR(COLOR_UPLOAD_BOX, FIELD_UPLOADS));

	mouse_clear();
	mouse_register(0, 0, max_x, DOWNLOAD_H, BUTTON1_PRESSED, mouse_listclick, &downloads.flat);
	mouse_register(0, 0, max_x, DOWNLOAD_H, BUTTON_UP, mouse_wheel, (void *) (0 | 2));
	mouse_register(0, 0, max_x, DOWNLOAD_H, BUTTON_DOWN, mouse_wheel, (void *) (1 | 2));
	mouse_register(0, DOWNLOAD_H + 1, max_x, UPLOAD_H, BUTTON1_PRESSED, mouse_listclick,
				   &uploads.flat);
	mouse_register(0, DOWNLOAD_H + 1, max_x, UPLOAD_H, BUTTON_UP, mouse_wheel, (void *) 0);
	mouse_register(0, DOWNLOAD_H + 1, max_x, UPLOAD_H, BUTTON_DOWN, mouse_wheel, (void *) 1);

	/* Predict the length of the help stuff on the bottom in order to center it. */
	if (len == -1)
		for (i = 0; help[i].key; i++)
			len += 1 + 3 + strlen(_(help[i].desc)) + HELP_PAD;

	x1 = len < max_x ? (max_x - len) / 2 : 0;

	for (i = 0; help[i].key; i++) {
		char buf[100];

		w1 = snprintf(buf, sizeof buf, "\\1%c\\0 - %s", help[i].key, _(help[i].desc));
		w1 -= 4;				/* do not count invisible characters */
		draw_fmt_str(x1, DOWNLOAD_H, max_x, 0, buf);
		mouse_register(x1, DOWNLOAD_H, w1, 1, BUTTON1_PRESSED, ui_mouse_simulate_keypress,
					   (void *) help[i].key);
		x1 += w1 + HELP_PAD;
	}

	if (formatted_xdim != max_x) {
		/* invalidate cached preformatted text */
		tree_touch_all(&downloads);
		tree_touch_all(&uploads);
		formatted_xdim = max_x;
	}

	transfer_screen_downloads_update();
	transfer_screen_uploads_update();
}

void transfer_screen_update_downloads_if_focus(void)
{
	if (active_screen == &transfer_screen_methods) {
		transfer_screen_downloads_update();
		refresh();
	}
}

void transfer_screen_update_uploads_if_focus(void)
{
	if (active_screen == &transfer_screen_methods) {
		transfer_screen_uploads_update();
		refresh();
	}
}

static void draw_transfers(tree * tr, int y, int height, int hilight, char *addentum)
{
	char buf[512] = "";
	char buf2[512];
	int i;
	int fwidth = max_x - 40;
	unsigned long received_sum = 0, total_sum = 0;
	unsigned long bandwidth_sum = 0;
	unsigned long total_eta = 0;

	if (fwidth < 8)
		fwidth = 8;

	for (i = 0; i < tr->top.num; i++) {
		transfer *t = list_index(&tr->top, i);

		if (transfer_alive(t)) {
			total_sum += t->filesize;
			if (t->bandwidth)
				total_eta += (t->filesize - t->transferred) / t->bandwidth;
			if (t->transferred != -1) {
				unsigned long bandwidth = t->bandwidth;

				received_sum += t->transferred;
				if (bandwidth != -1)
					bandwidth_sum += bandwidth;
			}
		}
	}
	tree_flatten(tr);
	list_check_values(&tr->flat, height - 3);
	draw_list_pretty(2, y, max_x - 4, height - 3, hilight, &tr->flat);

	if (total_sum)
		strcpy(buf, "\\5");
	/* Update the "total" line. */
	if (addentum)
		strncat(buf, addentum, sizeof buf - 4);
	strpad(buf, sizeof buf - 1);
	i = snprintf(buf2, sizeof buf2, _("Total: %s/%s at %sB/s"),
				 humanify(received_sum), humanify(total_sum), humanify(bandwidth_sum));
	if (bandwidth_sum)
		i += strncatf(buf2, sizeof buf2, " %6s", humanify_time(total_eta));
	i = max_x - 2 - i;
	strncpy(buf + i, buf2, sizeof buf - i);

	if (total_sum)
		insert_at_ratio(buf + 2, "\\0", total_sum ? (float) received_sum / total_sum : 0,
						max_x - 4);
	draw_fmt_str(2, y + height - 3, max_x - 4, 0, buf);
}

static void transfer_screen_downloads_update(void)
{
	char buf[512] = "";

#if HAVE_STATFS
	if (download_dir) {
		unsigned long free_kb = disk_free_kb();

		if (free_kb != -1)
			snprintf(buf, sizeof buf, _("Disk: %s free"), humanify_scale(free_kb, 1));
	}
#endif
	draw_transfers(&downloads, 1, DOWNLOAD_H, scr.active_field == FIELD_DOWNLOADS, buf);
}

static void transfer_screen_uploads_update(void)
{
	draw_transfers(&uploads, DOWNLOAD_H + 2, UPLOAD_H, scr.active_field == FIELD_UPLOADS, NULL);
}

static int transfer_screen_timer(void *data)
{
	if (active_screen == &transfer_screen_methods) {
		transfers_calculate();
		transfer_screen_downloads_update();
		transfer_screen_uploads_update();
		refresh();
	}
	return SECS(1);
}

#ifdef MOUSE
static void mouse_listclick(int rx, int ry, void *data)
{
	list *foo = data;
	int h = 0, prevfield = scr.active_field;

	if (active_screen != &transfer_screen_methods)
		return;

	if (foo == &downloads.flat) {
		h = DOWNLOAD_H;
		scr.active_field = FIELD_DOWNLOADS;
	} else if (foo == &uploads.flat) {
		h = UPLOAD_H;
		scr.active_field = FIELD_UPLOADS;
	} else
		assert(0);

	if (ry == 0)
		foo->sel -= h;
	else if (ry == h + 1)
		foo->sel += h;
	else
		foo->sel = foo->start + ry - 1;

	/* Expand/collapse a transfer result. */
	if (rx == 2)
		ui_handler('e');

	if (scr.active_field != prevfield)
		ui_draw();
	else if (data == &downloads.flat)
		transfer_screen_downloads_update();
	else
		transfer_screen_uploads_update();
	refresh();
}

static void mouse_wheel(int rx, int ry, void *data)
{
	int dir = (int) data;
	tree *foo;
	int h;

	if (active_screen != &transfer_screen_methods)
		return;

	if (dir & 2) {
		foo = &downloads;
		h = DOWNLOAD_H;
		scr.active_field = FIELD_DOWNLOADS;
	} else {
		foo = &uploads;
		h = UPLOAD_H;
		scr.active_field = FIELD_UPLOADS;
	}

	if (dir & ~2)
		foo->flat.sel += (h / 2) - 1;
	else
		foo->flat.sel -= (h / 2) - 1;

	list_check_values(&foo->flat, h - 1);
	ui_draw();
	refresh();
}
#endif

/* This function is used to create an entry in the lists in the transfer
 * screen. transfer_touch() is called to free t->pretty and set it to
 * NULL, thus we don't have to do anything if t->pretty is set.
 */
static char *pretty_transfer(transfer * t)
{
	if (t->pretty)
		return t->pretty;

	if (transfer_style == STYLE_ORIGINAL)
		return pretty_transfer_original(t);
	return pretty_transfer_jasta(t);
}

static char *pretty_source(source * s)
{
	transfer *t;

	if (s->pretty)
		return s->pretty;
	if (!(t = tree_find_parent(&downloads, s)))
		t = tree_find_parent(&uploads, s);
	assert(t);

	if (transfer_style == STYLE_ORIGINAL)
		return pretty_source_original(t, s);
	return pretty_source_jasta(t, s);
}

static list *transfer_expansion(transfer * t)
{
	return t->expanded ? &t->sourcen : NULL;
}

rendering transfer_methods = {
	(char *(*)(void *)) pretty_transfer,
	(list * (*)(void *)) transfer_expansion,
	(void (*)(void *)) transfer_destroy,
	(void (*)(void *)) transfer_touch,
};
rendering source_methods = {
	(char *(*)(void *)) pretty_source,
	NULL,
	(void (*)(void *)) source_destroy,
	(void (*)(void *)) source_touch,
};
ui_methods transfer_screen_methods = {
	transfer_screen_draw, transfer_screen_handler
};
