/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001-2002 Gran Weinholt <weinholt@linux.nu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: xcommon.c,v 1.41 2002/06/30 13:14:12 chnix Exp $
 */
#include "giftcurs.h"

#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "xcommon.h"
#include "screen.h"
#include "parse.h"

static void xputs(const char *s, int fd);

#ifdef HAVE_GETADDRINFO
int xconnect(char *hostname, char *port)
{
	struct addrinfo *res, *res0, hints = { 0, AF_UNSPEC, SOCK_STREAM };
	int sock = -1, e;

	if ((e = getaddrinfo(hostname, port, &hints, &res0)))
		DIE("getaddrinfo %s:%s: %s (%d)", hostname, port, gai_strerror(e), e);

	for (res = res0; res; res = res->ai_next) {
		sock = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
		if (sock < 0)
			continue;

		if (connect(sock, res->ai_addr, res->ai_addrlen) < 0) {
			close(sock);
			sock = -1;
			continue;
		}
		break;
	}
	freeaddrinfo(res0);

	if (sock < 0) {
		ERROR("socket||connect");
		return -1;
	}

	return sock;
}
#else
int xconnect(char *hostname, char *port)
{
	int sock;

	if (hostname[0] == '/') {
		/* This is a unix domain socket. */
		struct sockaddr_un saddr;

		if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
			FATAL("socket");

		saddr.sun_family = AF_UNIX;
		strncpy(saddr.sun_path, hostname, sizeof saddr.sun_path);
		if (connect(sock, (struct sockaddr *) &saddr, sizeof saddr) < 0) {
			ERROR("connect");
			close(sock);
			return -1;
		}
	} else {
		struct hostent *host;
		struct sockaddr_in saddr;
		struct servent *serv;

		if (!(serv = getservbyname(port, NULL)))
			DIE("%s: Unknown service", port);

		if (!(host = gethostbyname(hostname)))
			DIE("%s: %s (%d)", hostname, hstrerror(h_errno), h_errno);
		if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1)
			FATAL("socket");

		memcpy(&saddr.sin_addr.s_addr, host->h_addr_list[0], host->h_length);
		saddr.sin_family = AF_INET;
		saddr.sin_port = serv->s_port;	/* Already in network byteorder */
		if (connect(sock, (struct sockaddr *) &saddr, sizeof saddr) < 0) {
			ERROR("connect");
			close(sock);
			return -1;
		}
	}

	return sock;
}
#endif

int xgetlines(int fd, char *buf, int buflen, int (*handle) (char *buf))
{
	int n;
	char *wrk, *nl;

	wrk = strchr(buf, '\0');
	n = read(fd, wrk, buf + buflen - wrk - 1);
	if (n < 0) {
		ERROR("read");
		*buf = '\0';
		return -1;
	}
	if (n == 0) {
		if (wrk != buf)
			DEBUG("Incomplete last line on fd %d (%s.)", fd, buf);
		*buf = '\0';
		return 0;
	}
	wrk[n] = '\0';

	nl = strchr(wrk, '\n');
	if (nl) {
		char *pos = buf;

		do {
			*nl = '\0';
			if (handle(pos) < 0) {
				*buf = '\0';
				return -1;
			}
			pos = nl + 1;
			nl = strchr(pos, '\n');
		} while (nl);
		memmove(buf, pos, wrk + n - pos + 1);
	} else if (wrk + n >= buf + buflen - 1) {
		DEBUG("Line too long on fd %d (%s.)", fd, buf);
		*buf = '\0';
		return -1;
	}
	return n;
}

void xvprintf(int fd, const char *fmt, va_list ap)
{
	char *p = vsprintf_alloc(fmt, ap);

	if (!p)
		return;
	xputs(p, fd);
	free(p);
}

/* Note, this doesn't append a newline like puts() does. */
static void xputs(const char *s, int fd)
{
	int len = strlen(s);
	int n = write(fd, s, len);

	DEBUG("<= %s", s);
	if (n != len) {
		if (n < 0)
			ERROR("write");
		else
			DEBUG("Partial write! %d bytes lost.", len - n);
	}
}
